import logging

from .version import __version__

from .lib import material, spawn
from .render.overlay import Overlay, OverlayRegistry
from .core import config, agent, action
from .core.action import Action
from .core.agent import Agent, Scripted
from .core.env import Env
from .core.terrain import MapGenerator, Terrain

MOTD = rf'''      ___           ___           ___           ___
     /__/\         /__/\         /__/\         /  /\      Version {__version__:<8}
     \  \:\       |  |::\       |  |::\       /  /::\
      \  \:\      |  |:|:\      |  |:|:\     /  /:/\:\    An open source
  _____\__\:\   __|__|:|\:\   __|__|:|\:\   /  /:/  \:\   project originally
 /__/::::::::\ /__/::::| \:\ /__/::::| \:\ /__/:/ \__\:\  founded by Joseph Suarez
 \  \:\~~\~~\/ \  \:\~~\__\/ \  \:\~~\__\/ \  \:\ /  /:/  and formalized at OpenAI
  \  \:\  ~~~   \  \:\        \  \:\        \  \:\  /:/
   \  \:\        \  \:\        \  \:\        \  \:\/:/    Now developed and
    \  \:\        \  \:\        \  \:\        \  \::/     maintained at MIT in
     \__\/         \__\/         \__\/         \__\/      Phillip Isola's lab '''

__all__ = ['Env', 'config', 'agent', 'Agent', 'Scripted', 'MapGenerator', 'Terrain',
        'action', 'Action', 'material', 'spawn',
        'Overlay', 'OverlayRegistry']

try:
  __all__.append('OpenSkillRating')
except RuntimeError:
  logging.error('Warning: OpenSkill not installed. Ignore if you do not need this feature')



from nmmo.systems.skill import Skills
from nmmo.entity import entity
from nmmo.lib.log import EventCode

# pylint: disable=no-member
class Player(entity.Entity):
  def __init__(self, realm, pos, agent, resilient=False):
    super().__init__(realm, pos, agent.iden, agent.policy)

    self.agent    = agent
    self.immortal = realm.config.IMMORTAL
    self.resources.resilient = resilient

    # Scripted hooks
    self.target = None
    self.vision = 7

    # Logs
    self.buys                     = 0
    self.sells                    = 0
    self.ration_consumed          = 0
    self.poultice_consumed        = 0
    self.ration_level_consumed    = 0
    self.poultice_level_consumed  = 0

    # initialize skills with the base level
    self.skills = Skills(realm, self)
    if realm.config.PROGRESSION_SYSTEM_ENABLED:
      for skill in self.skills.skills:
        skill.level.update(realm.config.PROGRESSION_BASE_LEVEL)

    # Gold: initialize with 1 gold (EXCHANGE_BASE_GOLD).
    # If the base amount is more than 1, alss check the npc's init gold.
    if realm.config.EXCHANGE_SYSTEM_ENABLED:
      self.gold.update(realm.config.EXCHANGE_BASE_GOLD)

  @property
  def serial(self):
    return self.ent_id

  @property
  def is_player(self) -> bool:
    return True

  @property
  def level(self) -> int:
    # a player's level is the max of all skills
    # CHECK ME: the initial level is 1 because of Basic skills,
    #   which are harvesting food/water and don't progress
    return max(e.level.val for e in self.skills.skills)

  def apply_damage(self, dmg, style):
    super().apply_damage(dmg, style)
    self.skills.apply_damage(style)

  # TODO(daveey): The returns for this function are a mess
  def receive_damage(self, source, dmg):
    if self.immortal:
      return False

    # super().receive_damage returns True if self is alive after taking dmg
    if super().receive_damage(source, dmg):
      return True

    if not self.config.ITEM_SYSTEM_ENABLED:
      return False

    # starting from here, source receive gold & inventory items
    if self.config.EXCHANGE_SYSTEM_ENABLED and source is not None:
      if self.gold.val > 0:
        source.gold.increment(self.gold.val)
        self.realm.event_log.record(EventCode.EARN_GOLD, source, amount=self.gold.val)
        self.gold.update(0)

    # TODO: make source receive the highest-level items first
    #   because source cannot take it if the inventory is full
    item_list = list(self.inventory.items)
    self._np_random.shuffle(item_list)
    for item in item_list:
      self.inventory.remove(item)

      # if source is None or NPC, destroy the item
      if source.is_player:
        # inventory.receive() returns True if the item is received
        # if source doesn't have space, inventory.receive() destroys the item
        if source.inventory.receive(item):
          self.realm.event_log.record(EventCode.LOOT_ITEM, source, item=item)
      else:
        item.destroy()

    # CHECK ME: this is an empty function. do we still need this?
    self.skills.receive_damage(dmg)
    return False

  @property
  def equipment(self):
    return self.inventory.equipment

  def packet(self):
    data = super().packet()
    data['entID']     = self.ent_id
    data['resource']  = self.resources.packet()
    data['skills']    = self.skills.packet()
    data['inventory'] = self.inventory.packet()

    return data

  def update(self, realm, actions):
    '''Post-action update. Do not include history'''
    super().update(realm, actions)

    # Spawsn battle royale style death fog
    # Starts at 0 damage on the specified config tick
    # Moves in from the edges by 1 damage per tile per tick
    # So after 10 ticks, you take 10 damage at the edge and 1 damage
    # 10 tiles in, 0 damage in farther
    # This means all agents will be force killed around
    # MAP_CENTER / 2 + 100 ticks after spawning
    fog = self.config.PLAYER_DEATH_FOG
    if fog is not None and self.realm.tick >= fog:
      row, col = self.pos
      cent = self.config.MAP_BORDER + self.config.MAP_CENTER // 2

      # Distance from center of the map
      dist = max(abs(row - cent), abs(col - cent))

      # Safe final area
      if dist > self.config.PLAYER_DEATH_FOG_FINAL_SIZE:
        # Damage based on time and distance from center
        time_dmg = self.config.PLAYER_DEATH_FOG_SPEED * (self.realm.tick - fog + 1)
        dist_dmg = dist - self.config.MAP_CENTER // 2
        dmg = max(0, dist_dmg + time_dmg)
        self.receive_damage(None, dmg)

    if not self.alive:
      return

    self.resources.update()
    self.skills.update()



import functools
from typing import Any, Dict, List, Callable
from collections import defaultdict
from copy import copy, deepcopy
import dill

import gym
import numpy as np
from pettingzoo.utils.env import AgentID, ParallelEnv

import nmmo
from nmmo.core import realm
from nmmo.core.config import Default
from nmmo.core.observation import Observation
from nmmo.core.tile import Tile
from nmmo.entity.entity import Entity
from nmmo.systems.item import Item
from nmmo.task import task_api, task_spec
from nmmo.task.game_state import GameStateGenerator
from nmmo.lib import seeding

class Env(ParallelEnv):
  # Environment wrapper for Neural MMO using the Parallel PettingZoo API

  #pylint: disable=no-value-for-parameter
  def __init__(self,
               config: Default = nmmo.config.Default(),
               seed = None):
    self._np_random = None
    self._np_seed = None
    self._reset_required = True
    self.seed(seed)
    super().__init__()

    self.config = config
    self.realm = realm.Realm(config, self._np_random)
    self.obs = None
    self._dummy_obs = None

    self.possible_agents = list(range(1, config.PLAYER_N + 1))
    self._agents = None
    self._dead_agents = set()
    self._dead_this_tick = None
    self.scripted_agents = set()

    self._gamestate_generator = GameStateGenerator(self.realm, self.config)
    self.game_state = None
    # Default task: rewards 1 each turn agent is alive
    self.tasks = task_api.nmmo_default_task(self.possible_agents)
    self.agent_task_map = None
    self._dummy_task_embedding = np.zeros(self.config.TASK_EMBED_DIM, dtype=np.float16)

    # curriculum file path, if provided, should exist
    self.curriculum_file_path = config.CURRICULUM_FILE_PATH
    if self.curriculum_file_path is not None:
      # try to open the file to check if it exists
      with open(self.curriculum_file_path, 'rb') as f:
        curriculum = dill.load(f) # pylint: disable=unused-variable
      f.close()

  @functools.cached_property
  def _obs_space(self):
    def box(rows, cols):
      return gym.spaces.Box(
          low=-2**15, high=2**15-1,
          shape=(rows, cols),
          dtype=np.int16)
    def mask_box(length):
      return gym.spaces.Box(low=0, high=1, shape=(length,), dtype=np.int8)

    obs_space = {
      "CurrentTick": gym.spaces.Discrete(self.config.HORIZON+1),
      "AgentId": gym.spaces.Discrete(self.config.PLAYER_N+1),
      "Tile": box(self.config.MAP_N_OBS, Tile.State.num_attributes),
      "Entity": box(self.config.PLAYER_N_OBS, Entity.State.num_attributes),
      "Task": gym.spaces.Box(low=-2**15, high=2**15-1,
                             shape=(self.config.TASK_EMBED_DIM,),
                             dtype=np.float16),
    }

    if self.config.ITEM_SYSTEM_ENABLED:
      obs_space["Inventory"] = box(self.config.INVENTORY_N_OBS, Item.State.num_attributes)

    if self.config.EXCHANGE_SYSTEM_ENABLED:
      obs_space["Market"] = box(self.config.MARKET_N_OBS, Item.State.num_attributes)

    if self.config.PROVIDE_ACTION_TARGETS:
      mask_spec = deepcopy(self._atn_space)
      for atn_str in mask_spec:
        for arg_str in mask_spec[atn_str]:
          mask_spec[atn_str][arg_str] = mask_box(self._atn_space[atn_str][arg_str].n)
      obs_space["ActionTargets"] = mask_spec

    return gym.spaces.Dict(obs_space)

  # pylint: disable=method-cache-max-size-none
  @functools.lru_cache(maxsize=None)
  def observation_space(self, agent: AgentID):
    '''Neural MMO Observation Space

    Args:
        agent: Agent ID

    Returns:
        observation: gym.spaces object contained the structured observation
        for the specified agent.'''
    return self._obs_space

  @functools.cached_property
  def _atn_space(self):
    actions = {}
    for atn in sorted(nmmo.Action.edges(self.config)):
      if atn.enabled(self.config):
        actions[atn.__name__] = {}  # use the string key
        for arg in sorted(atn.edges):
          n = arg.N(self.config)
          actions[atn.__name__][arg.__name__] = gym.spaces.Discrete(n)
        actions[atn.__name__] = gym.spaces.Dict(actions[atn.__name__])
    return gym.spaces.Dict(actions)

  @functools.cached_property
  def _str_atn_map(self):
    '''Map action and argument names to their corresponding objects'''
    str_map = {}
    for atn in nmmo.Action.edges(self.config):
      str_map[atn.__name__] = atn
      for arg in atn.edges:
        str_map[arg.__name__] = arg
    return str_map

  # pylint: disable=method-cache-max-size-none
  @functools.lru_cache(maxsize=None)
  def action_space(self, agent: AgentID):
    '''Neural MMO Action Space

    Args:
        agent: Agent ID

    Returns:
        actions: gym.spaces object contained the structured actions
        for the specified agent. Each action is parameterized by a list
        of discrete-valued arguments. These consist of both fixed, k-way
        choices (such as movement direction) and selections from the
        observation space (such as targeting)'''
    return self._atn_space

  ############################################################################
  # Core API

  # TODO: This doesn't conform to the PettingZoo API
  # pylint: disable=arguments-renamed
  def reset(self, map_id=None, seed=None, options=None,
            make_task_fn: Callable=None):
    '''OpenAI Gym API reset function

    Loads a new game map and returns initial observations

    Args:
        map_id: Map index to load. Selects a random map by default
        seed: random seed to use
        make_task_fn: A function to make tasks

    Returns:
        observations, as documented by _compute_observations()

    Notes:
        Neural MMO simulates a persistent world. Ideally, you should reset
        the environment only once, upon creation. In practice, this approach
        limits the number of parallel environment simulations to the number
        of CPU cores available. At small and medium hardware scale, we
        therefore recommend the standard approach of resetting after a long
        but finite horizon: ~1000 timesteps for small maps and
        5000+ timesteps for large maps
    '''
    self.seed(seed)
    self.realm.reset(self._np_random, map_id)
    self._agents = list(self.realm.players.keys())
    self._dead_agents = set()
    self._dead_this_tick = {}

    # check if there are scripted agents
    for eid, ent in self.realm.players.items():
      if isinstance(ent.agent, nmmo.Scripted):
        self.scripted_agents.add(eid)
        ent.agent.set_rng(self._np_random)

    if self.curriculum_file_path is not None:
      self.tasks = self._sample_training_tasks()
    elif make_task_fn is not None:
      self.tasks = make_task_fn()
    else:
      for task in self.tasks:
        task.reset()
    self.agent_task_map = self._map_task_to_agent()

    self._dummy_obs = self._make_dummy_obs()
    self.obs = self._compute_observations()
    self._gamestate_generator = GameStateGenerator(self.realm, self.config)
    if self.game_state is not None:
      self.game_state.clear_cache()
      self.game_state = None

    self._reset_required = False

    return {a: o.to_gym() for a,o in self.obs.items()}

  def _sample_training_tasks(self):
    with open(self.curriculum_file_path, 'rb') as f:
      # curriculum file may have been changed, so read the file when sampling
      curriculum = dill.load(f) # a list of TaskSpec

    sampling_weights = [spec.sampling_weight for spec in curriculum]
    sampled_spec = self._np_random.choice(curriculum, size=len(self.possible_agents),
                                          p=sampling_weights/np.sum(sampling_weights))

    return task_spec.make_task_from_spec(self.possible_agents, sampled_spec)

  def _map_task_to_agent(self):
    agent_task_map: Dict[int, List[task_api.Task]] = {}
    for task in self.tasks:
      if task.embedding is None:
        task.set_embedding(self._dummy_task_embedding)
      # validate task embedding
      assert self._obs_space['Task'].contains(task.embedding), "Task embedding is not valid"

      # map task to agents
      for agent_id in task.assignee:
        if agent_id in agent_task_map:
          agent_task_map[agent_id].append(task)
        else:
          agent_task_map[agent_id] = [task]

    # for now we only support one task per agent
    if self.config.ALLOW_MULTI_TASKS_PER_AGENT is False:
      for agent_tasks in agent_task_map.values():
        assert len(agent_tasks) == 1, "Only one task per agent is supported"

    return agent_task_map

  def step(self, actions: Dict[int, Dict[str, Dict[str, Any]]]):
    '''Simulates one game tick or timestep

    Args:
        actions: A dictionary of agent decisions of format::

              {
                agent_1: {
                    action_1: [arg_1, arg_2],
                    action_2: [...],
                    ...
                },
                agent_2: {
                    ...
                },
                ...
              }

          Where agent_i is the integer index of the i\'th agent

          The environment only evaluates provided actions for provided
          gents. Unprovided action types are interpreted as no-ops and
          illegal actions are ignored

          It is also possible to specify invalid combinations of valid
          actions, such as two movements or two attacks. In this case,
          one will be selected arbitrarily from each incompatible sets.

          A well-formed algorithm should do none of the above. We only
          Perform this conditional processing to make batched action
          computation easier.

    Returns:
        (dict, dict, dict, None):

        observations:
          A dictionary of agent observations of format::

              {
                agent_1: obs_1,
                agent_2: obs_2,
                ...
              }

          Where agent_i is the integer index of the i\'th agent and
          obs_i is specified by the observation_space function.

        rewards:
          A dictionary of agent rewards of format::

              {
                agent_1: reward_1,
                agent_2: reward_2,
                ...
              }

          Where agent_i is the integer index of the i\'th agent and
          reward_i is the reward of the i\'th' agent.

          By default, agents receive -1 reward for dying and 0 reward for
          all other circumstances. Override Env.reward to specify
          custom reward functions

        dones:
          A dictionary of agent done booleans of format::

              {
                agent_1: done_1,
                agent_2: done_2,
                ...
              }

          Where agent_i is the integer index of the i\'th agent and
          done_i is a boolean denoting whether the i\'th agent has died.

          Note that obs_i will be a garbage placeholder if done_i is true.
          This is provided only for conformity with PettingZoo. Your
          algorithm should not attempt to leverage observations outside of
          trajectory bounds. You can omit garbage obs_i values by setting
          omitDead=True.

        infos:
          A dictionary of agent infos of format:

              {
                agent_1: None,
                agent_2: None,
                ...
              }

          Provided for conformity with PettingZoo
    '''
    assert not self._reset_required, 'step() called before reset'
    # Add in scripted agents' actions, if any
    if self.scripted_agents:
      actions = self._compute_scripted_agent_actions(actions)

    # Drop invalid actions of BOTH neural and scripted agents
    #   we don't need _deserialize_scripted_actions() anymore
    actions = self._validate_actions(actions)
    # Execute actions
    self._dead_this_tick = self.realm.step(actions)
    # the list of "current" agents, both alive and dead_this_tick
    self._agents = list(set(list(self.realm.players.keys()) + list(self._dead_this_tick.keys())))

    dones = {}
    for agent_id in self.agents:
      if agent_id in self._dead_this_tick or \
        self.realm.tick >= self.config.HORIZON or \
        (self.config.RESET_ON_DEATH and len(self._dead_agents) > 0):
        self._dead_agents.add(agent_id)
        dones[agent_id] = True
      else:
        dones[agent_id] = False

    # Store the observations, since actions reference them
    self.obs = self._compute_observations()
    gym_obs = {a: o.to_gym() for a,o in self.obs.items()}

    rewards, infos = self._compute_rewards()

    # NOTE: all obs, rewards, dones, infos have data for each agent in self.agents
    return gym_obs, rewards, dones, infos

  def _validate_actions(self, actions: Dict[int, Dict[str, Dict[str, Any]]]):
    '''Deserialize action arg values and validate actions
       For now, it does a basic validation (e.g., value is not none).
    '''
    validated_actions = {}

    for ent_id, atns in actions.items():
      if ent_id not in self.realm.players:
        #assert ent_id in self.realm.players, f'Entity {ent_id} not in realm'
        continue # Entity not in the realm -- invalid actions

      entity = self.realm.players[ent_id]
      if not entity.alive:
        #assert entity.alive, f'Entity {ent_id} is dead'
        continue # Entity is dead -- invalid actions

      validated_actions[ent_id] = {}

      for atn_key, args in sorted(atns.items()):
        action_valid = True
        deserialized_action = {}
        atn = self._str_atn_map[atn_key] if isinstance(atn_key, str) else atn_key
        if not atn.enabled(self.config):
          action_valid = False
          break

        for arg_key, val in sorted(args.items()):
          arg = self._str_atn_map[arg_key] if isinstance(arg_key, str) else arg_key
          obj = arg.deserialize(self.realm, entity, val, self.obs[ent_id])
          if obj is None:
            action_valid = False
            break
          deserialized_action[arg] = obj

        if action_valid:
          validated_actions[ent_id][atn] = deserialized_action

    return validated_actions

  def _compute_scripted_agent_actions(self, actions: Dict[int, Dict[str, Dict[str, Any]]]):
    '''Compute actions for scripted agents and add them into the action dict'''
    dead_agents = set()
    for agent_id in self.scripted_agents:
      if agent_id in self.realm.players:
        # override the provided scripted agents' actions
        actions[agent_id] = self.realm.players[agent_id].agent(self.obs[agent_id])
      else:
        dead_agents.add(agent_id)

    # remove the dead scripted agent from the list
    self.scripted_agents -= dead_agents

    return actions

  def _make_dummy_obs(self):
    dummy_tiles = np.zeros((0, len(Tile.State.attr_name_to_col)), dtype=np.int16)
    dummy_entities = np.zeros((0, len(Entity.State.attr_name_to_col)), dtype=np.int16)
    dummy_inventory = np.zeros((0, len(Item.State.attr_name_to_col)), dtype=np.int16)
    dummy_market = np.zeros((0, len(Item.State.attr_name_to_col)), dtype=np.int16)
    return Observation(self.config, self.realm.tick, 0, self._dummy_task_embedding,
                       dummy_tiles, dummy_entities, dummy_inventory, dummy_market)

  def _compute_observations(self):
    # Clean up unnecessary observations, which cause memory leaks
    if self.obs is not None:
      for agent_id, agent_obs in self.obs.items():
        agent_obs.clear_cache()  # clear the lru_cache
        self.obs[agent_id] = None
        del agent_obs
      self.obs = None

    obs = {}
    market = Item.Query.for_sale(self.realm.datastore)

    # get tile map, to bypass the expensive tile window query
    tile_map = Tile.Query.get_map(self.realm.datastore, self.config.MAP_SIZE)
    radius = self.config.PLAYER_VISION_RADIUS
    tile_obs_size = ((2*radius+1)**2, len(Tile.State.attr_name_to_col))

    for agent_id in self.agents:
      if agent_id not in self.realm.players:
        # return dummy obs for the agents in dead_this_tick
        dummy_obs = copy(self._dummy_obs)
        dummy_obs.current_tick = self.realm.tick
        dummy_obs.agent_id = agent_id
        obs[agent_id] = dummy_obs
      else:
        agent = self.realm.players.get(agent_id)
        agent_r = agent.row.val
        agent_c = agent.col.val

        visible_entities = Entity.Query.window(
            self.realm.datastore,
            agent_r, agent_c,
            radius
        )
        visible_tiles = tile_map[agent_r-radius:agent_r+radius+1,
                                 agent_c-radius:agent_c+radius+1,:].reshape(tile_obs_size)

        inventory = Item.Query.owned_by(self.realm.datastore, agent_id)

        # NOTE: the tasks for each agent is in self.agent_task_map, and task embeddings are
        #   available in each task instance, via task.embedding
        #   For now, each agent is assigned to a single task, so we just use the first task
        # TODO: can the embeddings of multiple tasks be superposed while preserving the
        #   task-specific information? This needs research
        task_embedding = self._dummy_task_embedding
        if agent_id in self.agent_task_map:
          task_embedding = self.agent_task_map[agent_id][0].embedding # NOTE: first task only
        obs[agent_id] = Observation(self.config, self.realm.tick, agent_id, task_embedding,
                                    visible_tiles, visible_entities, inventory, market)
    return obs

  def _compute_rewards(self):
    '''Computes the reward for the specified agent

    Override this method to create custom reward functions. You have full
    access to the environment state via self.realm. Our baselines do not
    modify this method; specify any changes when comparing to baselines

    Args:
        player: player object

    Returns:
        reward:
          The reward for the actions on the previous timestep of the
          entity identified by ent_id.
    '''
    # Initialization
    agents = set(self.agents)
    infos = {agent_id: {'task': {}} for agent_id in agents}
    rewards = defaultdict(int)

    # Clean up unnecessary game state, which cause memory leaks
    if self.game_state is not None:
      self.game_state.clear_cache()
      self.game_state = None

    # Compute Rewards and infos
    self.game_state = self._gamestate_generator.generate(self.realm, self.obs)
    for task in self.tasks:
      if agents.intersection(task.assignee): # evaluate only if the agents are current
        task_rewards, task_infos = task.compute_rewards(self.game_state)
        for agent_id, reward in task_rewards.items():
          if agent_id in agents:
            rewards[agent_id] = rewards.get(agent_id,0) + reward
            infos[agent_id]['task'][task.name] = task_infos[agent_id] # include progress, etc.
      else:
        task.close()  # To prevent memory leak

    # Make sure the dead agents return the rewards of -1
    for agent_id in self._dead_this_tick:
      rewards[agent_id] = -1

    return rewards, infos

  ############################################################################
  # PettingZoo API
  ############################################################################

  def render(self, mode='human'):
    '''For conformity with the PettingZoo API only; rendering is external'''

  @property
  def agents(self) -> List[AgentID]:
    '''For conformity with the PettingZoo API'''
    # returns the list of "current" agents, both alive and dead_this_tick
    return self._agents

  def close(self):
    '''For conformity with the PettingZoo API only; rendering is external'''

  def seed(self, seed=None):
    '''Reseeds the environment. reset() must be called after seed(), and before step().
       - self._np_seed is None: seed() has not been called, e.g. __init__() -> new RNG
       - self._np_seed is set, and seed is not None: seed() or reset() with seed -> new RNG

       If self._np_seed is set, but seed is None
         probably called from reset() without seed, so don't change the RNG
    '''
    if self._np_seed is None or seed is not None:
      self._np_random, self._np_seed = seeding.np_random(seed)
      self._reset_required = True

  def state(self) -> np.ndarray:
    raise NotImplementedError

  metadata = {'render.modes': ['human'], 'name': 'neural-mmo'}



import logging
import unittest

from copy import deepcopy
from timeit import timeit
import numpy as np

import nmmo
from nmmo.core import action
from nmmo.systems import item as Item
from nmmo.core.realm import Realm
from nmmo.lib import material as Material

from scripted import baselines

# this function can be replaced by assertDictEqual
# but might be still useful for debugging
def actions_are_equal(source_atn, target_atn, debug=True):

  # compare the numbers and player ids
  player_src = list(source_atn.keys())
  player_tgt = list(target_atn.keys())
  if player_src != player_tgt:
    if debug:
      logging.error("players don't match")
    return False

  # for each player, compare the actions
  for ent_id in player_src:
    atn1 = source_atn[ent_id]
    atn2 = target_atn[ent_id]

    if list(atn1.keys()) != list(atn2.keys()):
      if debug:
        logging.error("action keys don't match. player: %s", str(ent_id))
      return False

    for atn, args in atn1.items():
      if atn2[atn] != args:
        if debug:
          logging.error("action args don't match. player: %s, action: %s", str(ent_id), str(atn))
        return False

  return True


# this function CANNOT be replaced by assertDictEqual
def observations_are_equal(source_obs, target_obs, debug=True):

  keys_src = list(source_obs.keys())
  keys_obs = list(target_obs.keys())
  if keys_src != keys_obs:
    if debug:
      #print("entities don't match")
      logging.error("entities don't match")
    return False

  for k in keys_src:
    ent_src = source_obs[k]
    ent_tgt = target_obs[k]
    if list(ent_src.keys()) != list(ent_tgt.keys()):
      if debug:
        #print(f"entries don't match. key: {k}")
        logging.error("entries don't match. key: %s", str(k))
      return False

    obj = ent_src.keys()
    for o in obj:

      # ActionTargets causes a problem here, so skip it
      if o == "ActionTargets":
        continue

      obj_src = ent_src[o]
      obj_tgt = ent_tgt[o]
      if np.sum(obj_src != obj_tgt) > 0:
        if debug:
          #print(f"objects don't match. key: {k}, obj: {o}")
          logging.error("objects don't match. key: %s, obj: %s", str(k), str(o))
        return False

  return True


def player_total(env):
  return sum(ent.gold.val for ent in env.realm.players.values())


def count_actions(tick, actions):
  cnt_action = {}
  for atn in (action.Move, action.Attack, action.Sell, action.Use, action.Give, action.Buy):
    cnt_action[atn] = 0

  for ent_id in actions:
    for atn, _ in actions[ent_id].items():
      if atn in cnt_action:
        cnt_action[atn] += 1
      else:
        cnt_action[atn] = 1

  info_str = f"Tick: {tick}, acting agents: {len(actions)}, action counts " + \
             f"move: {cnt_action[action.Move]}, attack: {cnt_action[action.Attack]}, " + \
             f"sell: {cnt_action[action.Sell]}, use: {cnt_action[action.Move]}, " + \
             f"give: {cnt_action[action.Give]}, buy: {cnt_action[action.Buy]}"
  logging.info(info_str)

  return cnt_action


class ScriptedAgentTestConfig(nmmo.config.Small, nmmo.config.AllGameSystems):

  __test__ = False

  LOG_ENV = True

  LOG_MILESTONES = True
  LOG_EVENTS = False
  LOG_VERBOSE = False

  PLAYER_DEATH_FOG = 5

  PLAYERS = [
    baselines.Fisher, baselines.Herbalist,
    baselines.Prospector,baselines.Carver, baselines.Alchemist,
    baselines.Melee, baselines.Range, baselines.Mage]


# pylint: disable=abstract-method,duplicate-code
class ScriptedAgentTestEnv(nmmo.Env):
  '''
  EnvTest step() bypasses some differential treatments for scripted agents
  To do so, actions of scripted must be serialized using the serialize_actions function above
  '''
  __test__ = False

  def __init__(self, config: nmmo.config.Config, seed=None):
    super().__init__(config=config, seed=seed)

    # all agent must be scripted agents when using ScriptedAgentTestEnv
    for ent in self.realm.players.values():
      assert isinstance(ent.agent, baselines.Scripted), 'All agent must be scripted.'

    # this is to cache the actions generated by scripted policies
    self.actions = {}

  def reset(self, map_id=None, seed=None, options=None):
    self.actions = {}
    return super().reset(map_id=map_id, seed=seed, options=options)

  def _compute_scripted_agent_actions(self, actions):
    assert actions is not None, "actions must be provided, even it's {}"
    # if actions are not provided, generate actions using the scripted policy
    if actions == {}:
      for eid, ent in self.realm.players.items():
        actions[eid] = ent.agent(self.obs[eid])

      # cache the actions for replay before deserialization
      self.actions = deepcopy(actions)

    # if actions are provided, just run ent.agent() to set the RNG to the same state
    else:
      # NOTE: This is a hack to set the random number generator to the same state
      # since scripted agents also use RNG. Without this, the RNG is in different state,
      # and the env.step() does not give the same results in the deterministic replay.
      for eid, ent in self.realm.players.items():
        ent.agent(self.obs[eid])

    return actions


def change_spawn_pos(realm: Realm, ent_id: int, new_pos):
  # check if the position is valid
  assert realm.map.tiles[new_pos].habitable, "Given pos is not habitable."
  assert realm.entity(ent_id), "No such entity in the realm"

  entity = realm.entity(ent_id)
  old_pos = entity.pos
  realm.map.tiles[old_pos].remove_entity(ent_id)

  # set to new pos
  entity.row.update(new_pos[0])
  entity.col.update(new_pos[1])
  entity.spawn_pos = new_pos
  realm.map.tiles[new_pos].add_entity(entity)

def provide_item(realm: Realm, ent_id: int,
                 item: Item.Item, level: int, quantity: int):
  for _ in range(quantity):
    realm.players[ent_id].inventory.receive(
      item(realm, level=level))


# pylint: disable=invalid-name,protected-access
class ScriptedTestTemplate(unittest.TestCase):

  @classmethod
  def setUpClass(cls):
    # only use Combat agents
    cls.config = ScriptedAgentTestConfig()
    cls.config.PROVIDE_ACTION_TARGETS = True

    cls.config.PLAYERS = [baselines.Melee, baselines.Range, baselines.Mage]
    cls.config.PLAYER_N = 3
    #cls.config.IMMORTAL = True

    # set up agents to test ammo use
    cls.policy = { 1:'Melee', 2:'Range', 3:'Mage' }
    # 1 cannot hit 3, 2 can hit 1, 3 cannot hit 2
    cls.spawn_locs = { 1:(17, 17), 2:(17, 19), 3:(21, 21) }
    cls.ammo = { 1:Item.Whetstone, 2:Item.Arrow, 3:Item.Runes }
    cls.ammo_quantity = 2

    # items to provide
    cls.init_gold = 5
    cls.item_level = [0, 3] # 0 can be used, 3 cannot be used
    cls.item_sig = {}

  def _make_item_sig(self):
    item_sig = {}
    for ent_id, ammo in self.ammo.items():
      item_sig[ent_id] = []
      for item in [ammo, Item.Top, Item.Gloves, Item.Ration, Item.Potion]:
        for lvl in self.item_level:
          item_sig[ent_id].append((item, lvl))

    return item_sig

  def _setup_env(self, random_seed, check_assert=True, remove_immunity=False):
    """ set up a new env and perform initial checks """
    config = deepcopy(self.config)

    if remove_immunity:
      config.COMBAT_SPAWN_IMMUNITY = 0

    env = ScriptedAgentTestEnv(config, seed=random_seed)
    env.reset()

    # provide money for all
    for ent_id in env.realm.players:
      env.realm.players[ent_id].gold.update(self.init_gold)

    # provide items that are in item_sig
    self.item_sig = self._make_item_sig()
    for ent_id, items in self.item_sig.items():
      for item_sig in items:
        if item_sig[0] == self.ammo[ent_id]:
          provide_item(env.realm, ent_id, item_sig[0], item_sig[1], self.ammo_quantity)
        else:
          provide_item(env.realm, ent_id, item_sig[0], item_sig[1], 1)

    # teleport the players, if provided with specific locations
    for ent_id, pos in self.spawn_locs.items():
      change_spawn_pos(env.realm, ent_id, pos)

    # Change entire map to grass to become habitable and non-harvestable
    MS = env.config.MAP_SIZE
    for i in range(MS):
      for j in range(MS):
        tile = env.realm.map.tiles[i,j]
        tile.material = Material.Grass
        tile.material_id.update(Material.Grass.index)
        tile.state = Material.Grass(env.config)

    env.obs = env._compute_observations()

    if check_assert:
      self._check_default_asserts(env)

    return env

  def _check_ent_mask(self, ent_obs, atn, target_id):
    assert atn in [action.Give, action.GiveGold], "Invalid action"
    gym_obs = ent_obs.to_gym()
    mask = gym_obs["ActionTargets"][atn.__name__]["Target"][:ent_obs.entities.len] > 0

    return target_id in ent_obs.entities.ids[mask]

  def _check_inv_mask(self, ent_obs, atn, item_sig):
    assert atn in [action.Destroy, action.Give, action.Sell, action.Use], "Invalid action"
    gym_obs = ent_obs.to_gym()
    mask = gym_obs["ActionTargets"][atn.__name__]["InventoryItem"][:ent_obs.inventory.len] > 0
    inv_idx = ent_obs.inventory.sig(*item_sig)

    return ent_obs.inventory.id(inv_idx) in ent_obs.inventory.ids[mask]

  def _check_mkt_mask(self, ent_obs, item_id):
    gym_obs = ent_obs.to_gym()
    mask = gym_obs["ActionTargets"]["Buy"]["MarketItem"][:ent_obs.market.len] > 0

    return item_id in ent_obs.market.ids[mask]

  def _check_default_asserts(self, env):
    """ The below asserts are based on the hardcoded values in setUpClass()
        This should not run when different values were used
    """
    # check if the agents are in specified positions
    for ent_id, pos in self.spawn_locs.items():
      self.assertEqual(env.realm.players[ent_id].pos, pos)

    for ent_id, sig_list in self.item_sig.items():
      # ammo instances are in the datastore and global item registry (realm)
      inventory = env.obs[ent_id].inventory
      self.assertTrue(inventory.len == len(sig_list))
      for inv_idx in range(inventory.len):
        item_id = inventory.id(inv_idx)
        self.assertTrue(Item.ItemState.Query.by_id(env.realm.datastore, item_id) is not None)
        self.assertTrue(item_id in env.realm.items)

      for lvl in self.item_level:
        inv_idx = inventory.sig(self.ammo[ent_id], lvl)
        self.assertTrue(inv_idx is not None)
        self.assertEqual(self.ammo_quantity, # provided 2 ammos
          Item.ItemState.parse_array(inventory.values[inv_idx]).quantity)

      # check ActionTargets
      ent_obs = env.obs[ent_id]

      if env.config.ITEM_SYSTEM_ENABLED:
        # USE InventoryItem mask
        for item_sig in sig_list:
          if item_sig[1] == 0:
            # items that can be used
            self.assertTrue(self._check_inv_mask(ent_obs, action.Use, item_sig))
          else:
            # items that are too high to use
            self.assertFalse(self._check_inv_mask(ent_obs, action.Use, item_sig))

      if env.config.EXCHANGE_SYSTEM_ENABLED:
        # SELL InventoryItem mask
        for item_sig in sig_list:
          # the agent can sell anything now
          self.assertTrue(self._check_inv_mask(ent_obs, action.Sell, item_sig))

        # BUY MarketItem mask -- there is nothing on the market, so mask should be all 0
        self.assertTrue(len(env.obs[ent_id].market.ids) == 0)

  def _check_assert_make_action(self, env, atn, test_cond):
    assert atn in [action.Give, action.GiveGold, action.Buy], "Invalid action"
    actions = {}
    for ent_id, cond in test_cond.items():
      ent_obs = env.obs[ent_id]

      if atn in [action.Give, action.GiveGold]:
        # self should be always masked
        self.assertFalse(self._check_ent_mask(ent_obs, atn, ent_id))

        # check if the target is masked as expected
        self.assertEqual(
          cond['ent_mask'],
          self._check_ent_mask(ent_obs, atn, cond['tgt_id']),
          f"ent_id: {ent_id}, atn: {ent_id}, tgt_id: {cond['tgt_id']}"
        )

      if atn in [action.Give]:
        self.assertEqual(
          cond['inv_mask'],
          self._check_inv_mask(ent_obs, atn, cond['item_sig']),
          f"ent_id: {ent_id}, atn: {ent_id}, tgt_id: {cond['item_sig']}"
        )

      if atn in [action.Buy]:
        self.assertEqual(
          cond['mkt_mask'],
          self._check_mkt_mask(ent_obs, cond['item_id']),
          f"ent_id: {ent_id}, atn: {ent_id}, tgt_id: {cond['item_id']}"
        )

      # append the actions
      if atn == action.Give:
        actions[ent_id] = { action.Give: {
          action.InventoryItem: env.obs[ent_id].inventory.sig(*cond['item_sig']),
          action.Target: env.obs[ent_id].entities.index(cond['tgt_id']) } }

      elif atn == action.GiveGold:
        actions[ent_id] = { action.GiveGold:
          { action.Target: env.obs[ent_id].entities.index(cond['tgt_id']),
           action.Price: action.Price.index(cond['gold']) } }

      elif atn == action.Buy:
        mkt_idx = ent_obs.market.index(cond['item_id'])
        actions[ent_id] = { action.Buy: { action.MarketItem: mkt_idx } }

    return actions

# pylint: disable=unnecessary-lambda,bad-builtin
def profile_env_step(action_target=True, tasks=None, condition=None):
  config = nmmo.config.Default()
  config.PLAYERS = [baselines.Sleeper] # the scripted agents doing nothing
  config.IMMORTAL = True # otherwise the agents will die
  config.PROVIDE_ACTION_TARGETS = action_target
  env = nmmo.Env(config, seed=0)
  if tasks is None:
    tasks = []
  env.reset(seed=0, make_task_fn=lambda: tasks)
  for _ in range(3):
    env.step({})

  env.obs = env._compute_observations()
  obs = deepcopy(env.obs)

  test_func = [
    ('env.step({}):', lambda: env.step({})),
    ('env.realm.step():', lambda: env.realm.step({})),
    ('env._compute_observations():', lambda: env._compute_observations()),
    ('obs.to_gym(), ActionTarget:', lambda: {a: o.to_gym() for a,o in obs.items()}),
    ('env._compute_rewards():', lambda: env._compute_rewards())
  ]

  if condition:
    print('=== Test condition:', condition, '===')
  for name, func in test_func:
    print(' -', name, timeit(func, number=100, globals=globals()))



from collections import defaultdict
import unittest

from nmmo.datastore.serialized import SerializedState

# pylint: disable=no-member,unused-argument,unsubscriptable-object

FooState = SerializedState.subclass("FooState", [
  "a", "b", "col"
])

FooState.Limits = {
  "a": (-10, 10),
}

class MockDatastoreRecord():
  def __init__(self):
    self._data = defaultdict(lambda: 0)

  def get(self, name):
    return self._data[name]

  def update(self, name, value):
    self._data[name] = value

class MockDatastore():
  def create_record(self, name):
    return MockDatastoreRecord()

  def register_object_type(self, name, attributes):
    assert name == "FooState"
    assert attributes == ["a", "b", "col"]

class TestSerialized(unittest.TestCase):

  def test_serialized(self):
    state = FooState(MockDatastore(), FooState.Limits)

    # initial value = 0
    self.assertEqual(state.a.val, 0)

    # if given value is within the range, set to the value
    state.a.update(1)
    self.assertEqual(state.a.val, 1)

    # if given a lower value than the min, set to min
    a_min, a_max = FooState.Limits["a"]
    state.a.update(a_min - 100)
    self.assertEqual(state.a.val, a_min)

    # if given a higher value than the max, set to max
    state.a.update(a_max + 100)
    self.assertEqual(state.a.val, a_max)

if __name__ == '__main__':
  unittest.main()



import nmmo
from scripted.baselines import Random

def test_rollout():
  config = nmmo.config.Default()
  config.PLAYERS = [Random]

  env = nmmo.Env(config)
  env.reset()
  for _ in range(128):
    env.step({})

if __name__ == '__main__':
  test_rollout()



import unittest

import numpy as np

from nmmo.datastore.numpy_datastore import NumpyDatastore


class TestDatastore(unittest.TestCase):

  def testdatastore_record(self):
    datastore = NumpyDatastore()
    datastore.register_object_type("TestObject", 2)
    c1 = 0
    c2 = 1

    o = datastore.create_record("TestObject")
    self.assertEqual([o.get(c1), o.get(c2)], [0, 0])

    o.update(c1, 1)
    o.update(c2, 2)
    self.assertEqual([o.get(c1), o.get(c2)], [1, 2])

    np.testing.assert_array_equal(
      datastore.table("TestObject").get([o.id]),
      np.array([[1, 2]]))

    o2 = datastore.create_record("TestObject")
    o2.update(c2, 2)
    np.testing.assert_array_equal(
      datastore.table("TestObject").get([o.id, o2.id]),
      np.array([[1, 2], [0, 2]]))

    np.testing.assert_array_equal(
      datastore.table("TestObject").where_eq(c2, 2),
      np.array([[1, 2], [0, 2]]))

    o.delete()
    np.testing.assert_array_equal(
      datastore.table("TestObject").where_eq(c2, 2),
      np.array([[0, 2]]))

if __name__ == '__main__':
  unittest.main()




#pylint: disable=unused-argument

import logging
logging.basicConfig(level=logging.INFO, stream=None)

def pytest_benchmark_scale_unit(config, unit, benchmarks, best, worst, sort):
  if unit == 'seconds':
    prefix = 'millisec'
    scale = 1000
  elif unit == 'operations':
    prefix = ''
    scale = 1
  else:
    raise RuntimeError(f"Unexpected measurement unit {unit}")
  return prefix, scale



from __future__ import annotations

import abc

import numpy as np
from ordered_set import OrderedSet

from nmmo.lib import material
from nmmo.systems import combat
from nmmo.lib.log import EventCode

### Infrastructure ###
class ExperienceCalculator:
  def __init__(self, config):
    if not config.PROGRESSION_SYSTEM_ENABLED:
      return
    self.config = config
    self.exp_threshold = np.array(config.PROGRESSION_EXP_THRESHOLD)
    assert len(self.exp_threshold) >= config.PROGRESSION_LEVEL_MAX,\
      "PROGRESSION_LEVEL_BY_EXP must have at least PROGRESSION_LEVEL_MAX entries"
    self.max_exp = self.exp_threshold[self.config.PROGRESSION_LEVEL_MAX - 1]

  def exp_at_level(self, level):
    level = min(max(level, self.config.PROGRESSION_BASE_LEVEL),
                self.config.PROGRESSION_LEVEL_MAX)
    return int(self.exp_threshold[level - 1])

  def level_at_exp(self, exp):
    if exp >= self.max_exp:
      return self.config.PROGRESSION_LEVEL_MAX
    return np.argmin(exp >= self.exp_threshold)

class SkillGroup:
  def __init__(self, realm, entity):
    self.config  = realm.config
    self.realm   = realm
    self.entity = entity

    self.experience_calculator = ExperienceCalculator(self.config)
    self.skills  = OrderedSet() # critical for determinism

  def update(self):
    for skill in self.skills:
      skill.update()

  def packet(self):
    data = {}
    for skill in self.skills:
      data[skill.__class__.__name__.lower()] = skill.packet()

    return data

class Skill(abc.ABC):
  def __init__(self, skill_group: SkillGroup):
    self.realm = skill_group.realm
    self.config = skill_group.config
    self.entity = skill_group.entity

    self.experience_calculator = skill_group.experience_calculator
    self.skill_group = skill_group
    skill_group.skills.add(self)

  def packet(self):
    data = {}
    data['exp']   = self.exp.val
    data['level'] = self.level.val
    return data

  def add_xp(self, xp):
    self.exp.increment(xp)
    new_level = int(self.experience_calculator.level_at_exp(self.exp.val))

    if new_level > self.level.val:
      self.level.update(new_level)
      self.realm.event_log.record(EventCode.LEVEL_UP, self.entity,
                                  skill=self, level=new_level)

      self.realm.log_milestone(f'Level_{self.__class__.__name__}', new_level,
        f"PROGRESSION: Reached level {new_level} {self.__class__.__name__}",
        tags={"player_id": self.entity.ent_id})

  def set_experience_by_level(self, level):
    self.exp.update(self.experience_calculator.level_at_exp(level))
    self.level.update(int(level))

  @property
  def level(self):
    raise NotImplementedError(f"Skill {self.__class__.__name__} "\
      "does not implement 'level' property")

  @property
  def exp(self):
    raise NotImplementedError(f"Skill {self.__class__.__name__} "\
      "does not implement 'exp' property")

### Skill Bases ###
class CombatSkill(Skill):
  def update(self):
    pass

class NonCombatSkill(Skill):
  def __init__(self, skill_group: SkillGroup):
    super().__init__(skill_group)
    self._dummy_value = DummyValue()  # for water and food

  @property
  def level(self):
    return self._dummy_value

  @property
  def exp(self):
    return self._dummy_value

class HarvestSkill(NonCombatSkill):
  def process_drops(self, matl, drop_table):
    if not self.config.ITEM_SYSTEM_ENABLED:
      return

    entity = self.entity

    # harvest without tool will only yield level-1 item even with high skill level
    # for example, fishing level=5 without rod will only yield level-1 ration
    level = 1
    tool  = entity.equipment.held
    if matl.tool is not None and isinstance(tool.item, matl.tool):
      level = min(1+tool.item.level.val, self.config.PROGRESSION_LEVEL_MAX)

    #TODO: double-check drop table quantity
    for drop in drop_table.roll(self.realm, level):
      assert drop.level.val == level, 'Drop level does not match roll specification'

      self.realm.log_milestone(f'Gather_{drop.__class__.__name__}',
        level, f"PROFESSION: Gathered level {level} {drop.__class__.__name__} "
        f"(level {self.level.val} {self.__class__.__name__})",
        tags={"player_id": entity.ent_id})

      if entity.inventory.space:
        entity.inventory.receive(drop)
        self.realm.event_log.record(EventCode.HARVEST_ITEM, entity, item=drop)

  def harvest(self, matl, deplete=True):
    entity = self.entity
    realm  = self.realm

    r, c = entity.pos
    if realm.map.tiles[r, c].state != matl:
      return False

    drop_table = realm.map.harvest(r, c, deplete)
    if drop_table:
      self.process_drops(matl, drop_table)

    return drop_table

  def harvest_adjacent(self, matl, deplete=True):
    entity = self.entity
    realm  = self.realm

    r, c      = entity.pos
    drop_table = None

    if realm.map.tiles[r-1, c].state == matl:
      drop_table = realm.map.harvest(r-1, c, deplete)
    if realm.map.tiles[r+1, c].state == matl:
      drop_table = realm.map.harvest(r+1, c, deplete)
    if realm.map.tiles[r, c-1].state == matl:
      drop_table = realm.map.harvest(r, c-1, deplete)
    if realm.map.tiles[r, c+1].state == matl:
      drop_table = realm.map.harvest(r, c+1, deplete)

    if drop_table:
      self.process_drops(matl, drop_table)

    return drop_table

class AmmunitionSkill(HarvestSkill):
  def process_drops(self, matl, drop_table):
    super().process_drops(matl, drop_table)
    if self.config.PROGRESSION_SYSTEM_ENABLED:
      self.add_xp(self.config.PROGRESSION_AMMUNITION_XP_SCALE)


class ConsumableSkill(HarvestSkill):
  def process_drops(self, matl, drop_table):
    super().process_drops(matl, drop_table)
    if self.config.PROGRESSION_SYSTEM_ENABLED:
      self.add_xp(self.config.PROGRESSION_CONSUMABLE_XP_SCALE)


### Skill groups ###
class Basic(SkillGroup):
  def __init__(self, realm, entity):
    super().__init__(realm, entity)

    self.water = Water(self)
    self.food  = Food(self)

  @property
  def basic_level(self):
    return 0.5 * (self.water.level
            + self.food.level)

class Harvest(SkillGroup):
  def __init__(self, realm, entity):
    super().__init__(realm, entity)

    self.fishing      = Fishing(self)
    self.herbalism    = Herbalism(self)
    self.prospecting  = Prospecting(self)
    self.carving      = Carving(self)
    self.alchemy      = Alchemy(self)

  @property
  def harvest_level(self):
    return max(self.fishing.level,
                self.herbalism.level,
                self.prospecting.level,
                self.carving.level,
                self.alchemy.level)

class Combat(SkillGroup):
  def __init__(self, realm, entity):
    super().__init__(realm, entity)

    self.melee = Melee(self)
    self.range = Range(self)
    self.mage  = Mage(self)

  def packet(self):
    data          = super().packet()
    data['level'] = combat.level(self)

    return data

  @property
  def combat_level(self):
    return max(self.melee.level,
                self.range.level,
                self.mage.level)

  def apply_damage(self, style):
    if self.config.PROGRESSION_SYSTEM_ENABLED:
      skill  = self.__dict__[style]
      skill.add_xp(self.config.PROGRESSION_COMBAT_XP_SCALE)

  def receive_damage(self, dmg):
    pass

class Skills(Basic, Harvest, Combat):
  pass

### Combat Skills ###
class Melee(CombatSkill):
  SKILL_ID = 1

  @property
  def level(self):
    return self.entity.melee_level

  @property
  def exp(self):
    return self.entity.melee_exp

class Range(CombatSkill):
  SKILL_ID = 2

  @property
  def level(self):
    return self.entity.range_level

  @property
  def exp(self):
    return self.entity.range_exp

class Mage(CombatSkill):
  SKILL_ID = 3

  @property
  def level(self):
    return self.entity.mage_level

  @property
  def exp(self):
    return self.entity.mage_exp

Melee.weakness = Mage
Range.weakness = Melee
Mage.weakness  = Range


### Basic/Harvest Skills ###

class DummyValue:
  def __init__(self, val=0):
    self.val = val

  def update(self, val):
    self.val = val

class Water(HarvestSkill):
  def update(self):
    config = self.config
    if not config.RESOURCE_SYSTEM_ENABLED:
      return

    if config.IMMORTAL:
      return

    depletion = config.RESOURCE_DEPLETION_RATE
    water = self.entity.resources.water
    water.decrement(depletion)

    if not self.harvest_adjacent(material.Water, deplete=False):
      return

    restore = np.floor(config.RESOURCE_BASE
                      * config.RESOURCE_HARVEST_RESTORE_FRACTION)
    water.increment(restore)

    self.realm.event_log.record(EventCode.DRINK_WATER, self.entity)

class Food(HarvestSkill):
  def update(self):
    config = self.config
    if not config.RESOURCE_SYSTEM_ENABLED:
      return

    if config.IMMORTAL:
      return

    depletion = config.RESOURCE_DEPLETION_RATE
    food = self.entity.resources.food
    food.decrement(depletion)

    if not self.harvest(material.Foilage):
      return

    restore = np.floor(config.RESOURCE_BASE
                      * config.RESOURCE_HARVEST_RESTORE_FRACTION)
    food.increment(restore)

    self.realm.event_log.record(EventCode.EAT_FOOD, self.entity)

class Fishing(ConsumableSkill):
  SKILL_ID = 4

  @property
  def level(self):
    return self.entity.fishing_level

  @property
  def exp(self):
    return self.entity.fishing_exp

  def update(self):
    self.harvest_adjacent(material.Fish)

class Herbalism(ConsumableSkill):
  SKILL_ID = 5

  @property
  def level(self):
    return self.entity.herbalism_level

  @property
  def exp(self):
    return self.entity.herbalism_exp

  def update(self):
    self.harvest(material.Herb)

class Prospecting(AmmunitionSkill):
  SKILL_ID = 6

  @property
  def level(self):
    return self.entity.prospecting_level

  @property
  def exp(self):
    return self.entity.prospecting_exp

  def update(self):
    self.harvest(material.Ore)

class Carving(AmmunitionSkill):
  SKILL_ID = 7

  @property
  def level(self):
    return self.entity.carving_level

  @property
  def exp(self):
    return self.entity.carving_exp

  def update(self,):
    self.harvest(material.Tree)

class Alchemy(AmmunitionSkill):
  SKILL_ID = 8

  @property
  def level(self):
    return self.entity.alchemy_level

  @property
  def exp(self):
    return self.entity.alchemy_exp

  def update(self):
    self.harvest(material.Crystal)



# pylint: disable=invalid-name, attribute-defined-outside-init, no-member
from typing import Dict
from collections import defaultdict

import nmmo
from nmmo import material
from nmmo.systems import skill
import nmmo.systems.item as item_system
from nmmo.lib import colors
from nmmo.core import action
from nmmo.core.observation import Observation

from scripted import attack, move


class Scripted(nmmo.Scripted):
  '''Template class for baseline scripted models.

  You may either subclass directly or mirror the __call__ function'''
  scripted = True
  color    = colors.Neon.SKY
  def __init__(self, config, idx):
    '''
    Args:
        config : A forge.blade.core.Config object or subclass object
    '''
    super().__init__(config, idx)
    self.health_max = config.PLAYER_BASE_HEALTH

    if config.RESOURCE_SYSTEM_ENABLED:
      self.food_max   = config.RESOURCE_BASE
      self.water_max  = config.RESOURCE_BASE

    self.spawnR    = None
    self.spawnC    = None

  @property
  def policy(self):
    return self.__class__.__name__

  @property
  def forage_criterion(self) -> bool:
    '''Return true if low on food or water'''
    min_level = 7 * self.config.RESOURCE_DEPLETION_RATE
    return self.me.food <= min_level or self.me.water <= min_level

  def forage(self):
    '''Min/max food and water using Dijkstra's algorithm'''
    # TODO: do not access realm._np_random directly. ALSO see below for all other uses
    move.forageDijkstra(self.config, self.ob, self.actions,
                        self.food_max, self.water_max, self._np_random)

  def gather(self, resource):
    '''BFS search for a particular resource'''
    return move.gatherBFS(self.config, self.ob, self.actions, resource, self._np_random)

  def explore(self):
    '''Route away from spawn'''
    move.explore(self.config, self.ob, self.actions, self.me.row, self.me.col, self._np_random)

  @property
  def downtime(self):
    '''Return true if agent is not occupied with a high-priority action'''
    return not self.forage_criterion and self.attacker is None

  def evade(self):
    '''Target and path away from an attacker'''
    move.evade(self.config, self.ob, self.actions, self.attacker, self._np_random)
    self.target     = self.attacker
    self.targetID   = self.attackerID
    self.targetDist = self.attackerDist

  def attack(self):
    '''Attack the current target'''
    if self.target is not None:
      assert self.targetID is not None
      style = self._np_random.choice(self.style)
      attack.target(self.config, self.actions, style, self.targetID)

  def target_weak(self): # pylint: disable=inconsistent-return-statements
    '''Target the nearest agent if it is weak'''
    if self.closest is None:
      return False

    selfLevel  = self.me.level
    targLevel  = max(self.closest.melee_level, self.closest.range_level, self.closest.mage_level)

    if self.closest.npc_type == 1 or \
       targLevel <= selfLevel <= 5 or \
       selfLevel >= targLevel + 3:
      self.target     = self.closest
      self.targetID   = self.closestID
      self.targetDist = self.closestDist

  def scan_agents(self):
    '''Scan the nearby area for agents'''
    self.closest, self.closestDist   = attack.closestTarget(self.config, self.ob)
    self.attacker, self.attackerDist = attack.attacker(self.config, self.ob)

    self.closestID = None
    if self.closest is not None:
      self.closestID = self.ob.entities.index(self.closest.id)

    self.attackerID = None
    if self.attacker is not None:
      self.attackerID = self.ob.entities.index(self.attacker.id)

    self.target     = None
    self.targetID   = None
    self.targetDist = None

  def adaptive_control_and_targeting(self, explore=True):
    '''Balanced foraging, evasion, and exploration'''
    self.scan_agents()

    if self.attacker is not None:
      self.evade()
      return

    if self.fog_criterion:
      self.explore()
    elif self.forage_criterion or not explore:
      self.forage()
    else:
      self.explore()

    self.target_weak()

  def process_inventory(self):
    if not self.config.ITEM_SYSTEM_ENABLED:
      return

    self.inventory   = {}
    self.best_items: Dict   = {}
    self.item_counts = defaultdict(int)

    self.item_levels = {
      item_system.Hat.ITEM_TYPE_ID: self.level,
      item_system.Top.ITEM_TYPE_ID: self.level,
      item_system.Bottom.ITEM_TYPE_ID: self.level,
      item_system.Spear.ITEM_TYPE_ID: self.me.melee_level,
      item_system.Bow.ITEM_TYPE_ID: self.me.range_level,
      item_system.Wand.ITEM_TYPE_ID: self.me.mage_level,
      item_system.Rod.ITEM_TYPE_ID: self.me.fishing_level,
      item_system.Gloves.ITEM_TYPE_ID: self.me.herbalism_level,
      item_system.Pickaxe.ITEM_TYPE_ID: self.me.prospecting_level,
      item_system.Axe.ITEM_TYPE_ID: self.me.carving_level,
      item_system.Chisel.ITEM_TYPE_ID: self.me.alchemy_level,
      item_system.Whetstone.ITEM_TYPE_ID: self.me.melee_level,
      item_system.Arrow.ITEM_TYPE_ID: self.me.range_level,
      item_system.Runes.ITEM_TYPE_ID: self.me.mage_level,
      item_system.Ration.ITEM_TYPE_ID: self.level,
      item_system.Potion.ITEM_TYPE_ID: self.level
    }

    for item_ary in self.ob.inventory.values:
      itm = item_system.ItemState.parse_array(item_ary)
      assert itm.quantity != 0

      # Too high level to equip or use
      if itm.type_id in self.item_levels and itm.level > self.item_levels[itm.type_id]:
        continue

      # cannot use listed item
      if itm.listed_price:
        continue

      self.item_counts[itm.type_id] += itm.quantity
      self.inventory[itm.id] = itm

      # Best by default
      if itm.type_id not in self.best_items:
        self.best_items[itm.type_id] = itm

      best_itm = self.best_items[itm.type_id]

      if itm.level > best_itm.level:
        self.best_items[itm.type_id] = itm

  def upgrade_heuristic(self, current_level, upgrade_level, price):
    return (upgrade_level - current_level) / max(price, 1)

  def process_market(self):
    if not self.config.EXCHANGE_SYSTEM_ENABLED:
      return

    self.market         = {}
    self.best_heuristic = {}

    for item_ary in self.ob.market.values:
      itm = item_system.ItemState.parse_array(item_ary)

      self.market[itm.id] = itm

      # Prune Unaffordable
      if itm.listed_price > self.me.gold:
        continue

      # Too high level to equip
      if itm.type_id in self.item_levels and itm.level > self.item_levels[itm.type_id] :
        continue

      #Current best item level
      current_level = 0
      if itm.type_id in self.best_items:
        current_level = self.best_items[itm.type_id].level

      itm.heuristic = self.upgrade_heuristic(current_level, itm.level, itm.listed_price)

      #Always count first item
      if itm.type_id not in self.best_heuristic:
        self.best_heuristic[itm.type_id] = itm
        continue

      #Better heuristic value
      if itm.heuristic > self.best_heuristic[itm.type_id].heuristic:
        self.best_heuristic[itm.type_id] = itm

  def equip(self, items: set):
    for type_id, itm in self.best_items.items():
      if type_id not in items:
        continue

      if itm.equipped or itm.listed_price:
        continue

      # InventoryItem needs where the item is (index) in the inventory
      self.actions[action.Use] = {
        action.InventoryItem: self.ob.inventory.index(itm.id)}

    return True

  def consume(self):
    if self.me.health <= self.health_max // 2 \
        and item_system.Potion.ITEM_TYPE_ID in self.best_items:
      itm = self.best_items[item_system.Potion.ITEM_TYPE_ID]
    elif (self.me.food == 0 or self.me.water == 0) \
        and item_system.Ration.ITEM_TYPE_ID in self.best_items:
      itm = self.best_items[item_system.Ration.ITEM_TYPE_ID]
    else:
      return

    if itm.listed_price:
      return

    # InventoryItem needs where the item is (index) in the inventory
    self.actions[action.Use] = {
      action.InventoryItem: self.ob.inventory.index(itm.id)}

  def sell(self, keep_k: dict, keep_best: set):
    for itm in self.inventory.values():
      price = int(max(itm.level, 1))
      assert itm.quantity > 0

      if itm.equipped or itm.listed_price:
        continue

      if itm.type_id in keep_k:
        owned = self.item_counts[itm.type_id]
        k     = keep_k[itm.type_id]
        if owned <= k:
          continue

      #Exists an equippable of the current class, best needs to be kept, and this is the best item
      if itm.type_id in self.best_items and \
        itm.type_id in keep_best and \
        itm.id == self.best_items[itm.type_id].id:
        continue

      self.actions[action.Sell] = {
        action.InventoryItem: self.ob.inventory.index(itm.id),
        action.Price: action.Price.index(price) }

      return itm

  def buy(self, buy_k: dict, buy_upgrade: set):
    if len(self.inventory) >= self.config.ITEM_INVENTORY_CAPACITY:
      return

    purchase = None
    best = list(self.best_heuristic.items())
    self._np_random.shuffle(best)
    for type_id, itm in best:
      # Buy top k
      if type_id in buy_k:
        owned = self.item_counts[type_id]
        k = buy_k[type_id]
        if owned < k:
          purchase = itm

      # Check if item desired and upgrade
      elif type_id in buy_upgrade and itm.heuristic > 0:
        purchase = itm

      # Buy best heuristic upgrade
      if purchase:
        self.actions[action.Buy] = {
          action.MarketItem: self.ob.market.index(purchase.id)}
        return

  def exchange(self):
    if not self.config.EXCHANGE_SYSTEM_ENABLED:
      return

    self.process_market()
    self.sell(keep_k=self.supplies, keep_best=self.wishlist)
    self.buy(buy_k=self.supplies, buy_upgrade=self.wishlist)

  def use(self):
    self.process_inventory()
    if self.config.EQUIPMENT_SYSTEM_ENABLED and not self.consume():
      self.equip(items=self.wishlist)

  def __call__(self, observation: Observation):
    '''Process observations and return actions'''
    assert self._np_random is not None, "Agent's RNG must be set."
    self.actions = {}

    self.ob = observation
    self.me = observation.agent()

    # combat level
    self.me.level = max(self.me.melee_level, self.me.range_level, self.me.mage_level)

    self.skills = {
      skill.Melee: self.me.melee_level,
      skill.Range: self.me.range_level,
      skill.Mage: self.me.mage_level,
      skill.Fishing: self.me.fishing_level,
      skill.Herbalism: self.me.herbalism_level,
      skill.Prospecting: self.me.prospecting_level,
      skill.Carving: self.me.carving_level,
      skill.Alchemy: self.me.alchemy_level
    }

    # TODO(kywch): need a consistent level variables
    # level for using armor, rations, and potion
    self.level = min(1, max(self.skills.values()))

    if self.spawnR is None:
      self.spawnR = self.me.row
    if self.spawnC is None:
      self.spawnC = self.me.col

    # When to run from death fog in BR configs
    self.fog_criterion = None
    if self.config.PLAYER_DEATH_FOG is not None:
      time_alive = self.me.time_alive
      start_running = time_alive > self.config.PLAYER_DEATH_FOG - 64
      run_now = time_alive % max(1, int(1 / self.config.PLAYER_DEATH_FOG_SPEED))
      self.fog_criterion = start_running and run_now


class Sleeper(Scripted):
  '''Do Nothing'''
  def __call__(self, obs):
    super().__call__(obs)
    return {}
class Random(Scripted):
  '''Moves randomly'''
  def __call__(self, obs):
    super().__call__(obs)

    move.rand(self.config, self.ob, self.actions, self._np_random)
    return self.actions

class Meander(Scripted):
  '''Moves randomly on safe terrain'''
  def __call__(self, obs):
    super().__call__(obs)

    move.meander(self.config, self.ob, self.actions, self._np_random)
    return self.actions

class Explore(Scripted):
  '''Actively explores towards the center'''
  def __call__(self, obs):
    super().__call__(obs)

    self.explore()

    return self.actions

class Forage(Scripted):
  '''Forages using Dijkstra's algorithm and actively explores'''
  def __call__(self, obs):
    super().__call__(obs)

    if self.forage_criterion:
      self.forage()
    else:
      self.explore()

    return self.actions

class Combat(Scripted):
  '''Forages, fights, and explores'''
  def __init__(self, config, idx):
    super().__init__(config, idx)
    self.style  = [action.Melee, action.Range, action.Mage]

  @property
  def supplies(self):
    return {
      item_system.Ration.ITEM_TYPE_ID: 2,
      item_system.Potion.ITEM_TYPE_ID: 2,
      self.ammo.ITEM_TYPE_ID: 10
    }

  @property
  def wishlist(self):
    return {
      item_system.Hat.ITEM_TYPE_ID,
      item_system.Top.ITEM_TYPE_ID,
      item_system.Bottom.ITEM_TYPE_ID,
      self.weapon.ITEM_TYPE_ID,
      self.ammo.ITEM_TYPE_ID
    }

  def __call__(self, obs):
    super().__call__(obs)
    self.use()
    self.exchange()

    self.adaptive_control_and_targeting()
    self.attack()

    return self.actions

class Gather(Scripted):
  '''Forages, fights, and explores'''
  def __init__(self, config, idx):
    super().__init__(config, idx)
    self.resource = [material.Fish, material.Herb, material.Ore, material.Tree, material.Crystal]

  @property
  def supplies(self):
    return {
      item_system.Ration.ITEM_TYPE_ID: 1,
      item_system.Potion.ITEM_TYPE_ID: 1
    }

  @property
  def wishlist(self):
    return {
      item_system.Hat.ITEM_TYPE_ID,
      item_system.Top.ITEM_TYPE_ID,
      item_system.Bottom.ITEM_TYPE_ID,
      self.tool.ITEM_TYPE_ID
    }

  def __call__(self, obs):
    super().__call__(obs)
    self.use()
    self.exchange()

    if self.forage_criterion:
      self.forage()
    elif self.fog_criterion or not self.gather(self.resource):
      self.explore()

    return self.actions

class Fisher(Gather):
  def __init__(self, config, idx):
    super().__init__(config, idx)
    if config.PROFESSION_SYSTEM_ENABLED:
      self.resource = [material.Fish]
    self.tool     = item_system.Rod

class Herbalist(Gather):
  def __init__(self, config, idx):
    super().__init__(config, idx)
    if config.PROFESSION_SYSTEM_ENABLED:
      self.resource = [material.Herb]
    self.tool     = item_system.Gloves

class Prospector(Gather):
  def __init__(self, config, idx):
    super().__init__(config, idx)
    if config.PROFESSION_SYSTEM_ENABLED:
      self.resource = [material.Ore]
    self.tool     = item_system.Pickaxe

class Carver(Gather):
  def __init__(self, config, idx):
    super().__init__(config, idx)
    if config.PROFESSION_SYSTEM_ENABLED:
      self.resource = [material.Tree]
    self.tool     = item_system.Axe

class Alchemist(Gather):
  def __init__(self, config, idx):
    super().__init__(config, idx)
    if config.PROFESSION_SYSTEM_ENABLED:
      self.resource = [material.Crystal]
    self.tool     = item_system.Chisel

class Melee(Combat):
  def __init__(self, config, idx):
    super().__init__(config, idx)
    if config.COMBAT_SYSTEM_ENABLED:
      self.style  = [action.Melee]
    self.weapon = item_system.Spear
    self.ammo   = item_system.Whetstone

class Range(Combat):
  def __init__(self, config, idx):
    super().__init__(config, idx)
    if config.COMBAT_SYSTEM_ENABLED:
      self.style  = [action.Range]
    self.weapon = item_system.Bow
    self.ammo   = item_system.Arrow

class Mage(Combat):
  def __init__(self, config, idx):
    super().__init__(config, idx)
    if config.COMBAT_SYSTEM_ENABLED:
      self.style  = [action.Mage]
    self.weapon = item_system.Wand
    self.ammo   = item_system.Runes



# pylint: disable=all

import inspect
from collections import deque

import numpy as np


class staticproperty(property):
  def __get__(self, cls, owner):
    return self.fget.__get__(None, owner)()

class classproperty(object):
  def __init__(self, f):
    self.f = f
  def __get__(self, obj, owner):
    return self.f(owner)

class Iterable(type):
  def __iter__(cls):
    queue = deque(cls.__dict__.items())
    while len(queue) > 0:
      name, attr = queue.popleft()
      if type(name) != tuple:
        name = tuple([name])
      if not inspect.isclass(attr):
        continue
      yield name, attr

  def values(cls):
    return [e[1] for e in cls]

class StaticIterable(type):
  def __iter__(cls):
    stack = list(cls.__dict__.items())
    stack.reverse()
    for name, attr in stack:
      if name == '__module__':
        continue
      if name.startswith('__'):
        break
      yield name, attr

class NameComparable(type):
  def __hash__(self):
    return hash(self.__name__)

  def __eq__(self, other):
    return self.__name__ == other.__name__

  def __ne__(self, other):
    return self.__name__ != other.__name__

  def __lt__(self, other):
    return self.__name__ < other.__name__

  def __le__(self, other):
    return self.__name__ <= other.__name__

  def __gt__(self, other):
    return self.__name__ > other.__name__

  def __ge__(self, other):
    return self.__name__ >= other.__name__

class IterableNameComparable(Iterable, NameComparable):
  pass

def linf(pos1, pos2):
  # pos could be a single (r,c) or a vector of (r,c)s
  diff = np.abs(np.array(pos1) - np.array(pos2))
  return np.max(diff, axis=-1)

def linf_single(pos1, pos2):
  # pos is a single (r,c) to avoid uneccessary function calls
  return max(abs(pos1[0]-pos2[0]), abs(pos1[1]-pos2[1]))

#Bounds checker
def in_bounds(r, c, shape, border=0):
  R, C = shape
  return (
    r > border and
    c > border and
    r < R - border and
    c < C - border
  )




# pylint: disable=invalid-name
from __future__ import annotations

import os
import sys
import logging

import nmmo
from nmmo.core.agent import Agent
from nmmo.core.terrain import MapGenerator
from nmmo.lib import utils, material, spawn

class Template(metaclass=utils.StaticIterable):
  def __init__(self):
    self.data = {}
    cls       = type(self)

    #Set defaults from static properties
    for k, v in cls:
      self.set(k, v)

  def override(self, **kwargs):
    for k, v in kwargs.items():
      err = f'CLI argument: {k} is not a Config property'
      assert hasattr(self, k), err
      self.set(k, v)

  def set(self, k, v):
    if not isinstance(v, property):
      try:
        setattr(self, k, v)
      except AttributeError:
        logging.error('Cannot set attribute: %s to %s', str(k), str(v))
        sys.exit()
    self.data[k] = v

  # pylint: disable=bad-builtin
  def print(self):
    key_len = 0
    for k in self.data:
      key_len = max(key_len, len(k))

    print('Configuration')
    for k, v in self.data.items():
      print(f'   {k:{key_len}s}: {v}')

  def items(self):
    return self.data.items()

  def __iter__(self):
    for k in self.data:
      yield k

  def keys(self):
    return self.data.keys()

  def values(self):
    return self.data.values()

def validate(config):
  err = 'config.Config is a base class. Use config.{Small, Medium Large}'''
  assert isinstance(config, Config), err

  if not config.TERRAIN_SYSTEM_ENABLED:
    err = 'Invalid Config: {} requires Terrain'
    assert not config.RESOURCE_SYSTEM_ENABLED, err.format('Resource')
    assert not config.PROFESSION_SYSTEM_ENABLED, err.format('Profession')

  if not config.COMBAT_SYSTEM_ENABLED:
    err = 'Invalid Config: {} requires Combat'
    assert not config.NPC_SYSTEM_ENABLED, err.format('NPC')

  if not config.ITEM_SYSTEM_ENABLED:
    err = 'Invalid Config: {} requires Inventory'
    assert not config.EQUIPMENT_SYSTEM_ENABLED, err.format('Equipment')
    assert not config.PROFESSION_SYSTEM_ENABLED, err.format('Profession')
    assert not config.EXCHANGE_SYSTEM_ENABLED, err.format('Exchange')


class Config(Template):
  '''An environment configuration object

  Global constants are defined as static class variables. You can override
  any Config variable using standard CLI syntax (e.g. --NENT=128).

  The default config as of v1.5 uses 1024x1024 maps with up to 2048 agents
  and 1024 NPCs. It is suitable to time horizons of 8192+ steps. For smaller
  experiments, consider the SmallMaps config.

  Notes:
    We use Google Fire internally to replace standard manual argparse
    definitions for each Config property. This means you can subclass
    Config to add new static attributes -- CLI definitions will be
    generated automatically.
  '''

  def __init__(self):
    super().__init__()

    # TODO: Come up with a better way
    # to resolve mixin MRO conflicts
    if not hasattr(self, 'TERRAIN_SYSTEM_ENABLED'):
      self.TERRAIN_SYSTEM_ENABLED = False

    if not hasattr(self, 'RESOURCE_SYSTEM_ENABLED'):
      self.RESOURCE_SYSTEM_ENABLED = False

    if not hasattr(self, 'COMBAT_SYSTEM_ENABLED'):
      self.COMBAT_SYSTEM_ENABLED = False

    if not hasattr(self, 'NPC_SYSTEM_ENABLED'):
      self.NPC_SYSTEM_ENABLED = False

    if not hasattr(self, 'PROGRESSION_SYSTEM_ENABLED'):
      self.PROGRESSION_SYSTEM_ENABLED = False

    if not hasattr(self, 'ITEM_SYSTEM_ENABLED'):
      self.ITEM_SYSTEM_ENABLED = False

    if not hasattr(self, 'EQUIPMENT_SYSTEM_ENABLED'):
      self.EQUIPMENT_SYSTEM_ENABLED = False

    if not hasattr(self, 'PROFESSION_SYSTEM_ENABLED'):
      self.PROFESSION_SYSTEM_ENABLED = False

    if not hasattr(self, 'EXCHANGE_SYSTEM_ENABLED'):
      self.EXCHANGE_SYSTEM_ENABLED = False

    if not hasattr(self, 'COMMUNICATION_SYSTEM_ENABLED'):
      self.COMMUNICATION_SYSTEM_ENABLED = False

    if __debug__:
      validate(self)

    deprecated_attrs = [
      'NENT', 'NPOP', 'AGENTS', 'NMAPS', 'FORCE_MAP_GENERATION', 'SPAWN']

    for attr in deprecated_attrs:
      assert not hasattr(self, attr), f'{attr} has been deprecated or renamed'


  ############################################################################
  ### Meta-Parameters
  def game_system_enabled(self, name) -> bool:
    return hasattr(self, name)

  PROVIDE_ACTION_TARGETS       = True
  '''Provide action targets mask'''

  PROVIDE_NOOP_ACTION_TARGET   = True
  '''Provide a no-op option for each action'''

  PLAYERS                      = [Agent]
  '''Player classes from which to spawn'''

  HORIZON = 1024
  '''Number of steps before the environment resets'''

  CURRICULUM_FILE_PATH = None
  '''Path to a curriculum task file containing a list of task specs for training'''

  TASK_EMBED_DIM = 4096
  '''Dimensionality of task embeddings'''

  ALLOW_MULTI_TASKS_PER_AGENT = False
  '''Whether to allow multiple tasks per agent'''

  ############################################################################
  ### Population Parameters
  LOG_VERBOSE                  = False
  '''Whether to log server messages or just stats'''

  LOG_ENV                      = False
  '''Whether to log env steps (expensive)'''

  LOG_MILESTONES               = True
  '''Whether to log server-firsts (semi-expensive)'''

  LOG_EVENTS                   = True
  '''Whether to log events (semi-expensive)'''

  LOG_FILE                     = None
  '''Where to write logs (defaults to console)'''


  ############################################################################
  ### Player Parameters
  PLAYER_N                     = None
  '''Maximum number of players spawnable in the environment'''

  # TODO: CHECK if there could be 100+ entities within one's vision
  PLAYER_N_OBS                 = 100
  '''Number of distinct agent observations'''

  @property
  def PLAYER_POLICIES(self):
    '''Number of player policies'''
    return len(self.PLAYERS)

  PLAYER_BASE_HEALTH           = 100
  '''Initial agent health'''

  PLAYER_VISION_RADIUS         = 7
  '''Number of tiles an agent can see in any direction'''

  @property
  def PLAYER_VISION_DIAMETER(self):
    '''Size of the square tile crop visible to an agent'''
    return 2*self.PLAYER_VISION_RADIUS + 1

  PLAYER_DEATH_FOG             = None
  '''How long before spawning death fog. None for no death fog'''

  PLAYER_DEATH_FOG_SPEED       = 1
  '''Number of tiles per tick that the fog moves in'''

  PLAYER_DEATH_FOG_FINAL_SIZE  = 8
  '''Number of tiles from the center that the fog stops'''

  PLAYER_LOADER                = spawn.SequentialLoader
  '''Agent loader class specifying spawn sampling'''

  PLAYER_SPAWN_TEAMMATE_DISTANCE = 1
  '''Buffer tiles between teammates at spawn'''

  @property
  def PLAYER_TEAM_SIZE(self):
    if __debug__:
      assert not self.PLAYER_N % len(self.PLAYERS)
    return self.PLAYER_N // len(self.PLAYERS)

  ############################################################################
  ### Debug Parameters
  IMMORTAL = False
  '''Debug parameter: prevents agents from dying except by void'''

  RESET_ON_DEATH = False
  '''Debug parameter: whether to reset the environment whenever an agent dies'''

  ############################################################################
  ### Map Parameters
  MAP_N                        = 1
  '''Number of maps to generate'''

  MAP_N_TILE                   = len(material.All.materials)
  '''Number of distinct terrain tile types'''

  @property
  def MAP_N_OBS(self):
    '''Number of distinct tile observations'''
    return int(self.PLAYER_VISION_DIAMETER ** 2)

  MAP_CENTER                   = None
  '''Size of each map (number of tiles along each side)'''

  MAP_BORDER                   = 16
  '''Number of void border tiles surrounding each side of the map'''

  @property
  def MAP_SIZE(self):
    return int(self.MAP_CENTER + 2*self.MAP_BORDER)

  MAP_GENERATOR                = MapGenerator
  '''Specifies a user map generator. Uses default generator if unspecified.'''

  MAP_FORCE_GENERATION         = True
  '''Whether to regenerate and overwrite existing maps'''

  MAP_GENERATE_PREVIEWS        = False
  '''Whether map generation should also save .png previews (slow + large file size)'''

  MAP_PREVIEW_DOWNSCALE        = 1
  '''Downscaling factor for png previews'''


  ############################################################################
  ### Path Parameters
  PATH_ROOT                = os.path.dirname(nmmo.__file__)
  '''Global repository directory'''

  PATH_CWD                 = os.getcwd()
  '''Working directory'''

  PATH_RESOURCE            = os.path.join(PATH_ROOT, 'resource')
  '''Resource directory'''

  PATH_TILE                = os.path.join(PATH_RESOURCE, '{}.png')
  '''Tile path -- format me with tile name'''

  PATH_MAPS                = None
  '''Generated map directory'''

  PATH_MAP_SUFFIX          = 'map{}/map.npy'
  '''Map file name'''

  PATH_MAP_SUFFIX          = 'map{}/map.npy'
  '''Map file name'''


############################################################################
### Game Systems (Static Mixins)
class Terrain:
  '''Terrain Game System'''

  TERRAIN_SYSTEM_ENABLED       = True
  '''Game system flag'''

  TERRAIN_FLIP_SEED            = False
  '''Whether to negate the seed used for generation (useful for unique heldout maps)'''

  TERRAIN_FREQUENCY            = -3
  '''Base noise frequency range (log2 space)'''

  TERRAIN_FREQUENCY_OFFSET     = 7
  '''Noise frequency octave offset (log2 space)'''

  TERRAIN_LOG_INTERPOLATE_MIN  = -2
  '''Minimum interpolation log-strength for noise frequencies'''

  TERRAIN_LOG_INTERPOLATE_MAX  = 0
  '''Maximum interpolation log-strength for noise frequencies'''

  TERRAIN_TILES_PER_OCTAVE     = 8
  '''Number of octaves sampled from log2 spaced TERRAIN_FREQUENCY range'''

  TERRAIN_VOID                 = 0.0
  '''Noise threshold for void generation'''

  TERRAIN_WATER                = 0.30
  '''Noise threshold for water generation'''

  TERRAIN_GRASS                = 0.70
  '''Noise threshold for grass'''

  TERRAIN_FOILAGE              = 0.85
  '''Noise threshold for foilage (food tile)'''


class Resource:
  '''Resource Game System'''

  RESOURCE_SYSTEM_ENABLED             = True
  '''Game system flag'''

  RESOURCE_BASE                       = 100
  '''Initial level and capacity for food and water'''

  RESOURCE_DEPLETION_RATE             = 5
  '''Depletion rate for food and water'''

  RESOURCE_STARVATION_RATE            = 10
  '''Damage per tick without food'''

  RESOURCE_DEHYDRATION_RATE           = 10
  '''Damage per tick without water'''

  RESOURCE_RESILIENT_POPULATION       = 0
  '''Training helper: proportion of population that is resilient to starvation and dehydration
     (e.g. 0.1 means 10% of the population is resilient to starvation and dehydration)
     This is to make some agents live longer during training to sample from "advanced" agents.'''

  RESOURCE_DAMAGE_REDUCTION           = 0.5
  '''Training helper: damage reduction from starvation and dehydration for resilient agents'''

  RESOURCE_FOILAGE_CAPACITY           = 1
  '''Maximum number of harvests before a foilage tile decays'''

  RESOURCE_FOILAGE_RESPAWN            = 0.025
  '''Probability that a harvested foilage tile will regenerate each tick'''

  RESOURCE_HARVEST_RESTORE_FRACTION   = 1.0
  '''Fraction of maximum capacity restored upon collecting a resource'''

  RESOURCE_HEALTH_REGEN_THRESHOLD     = 0.5
  '''Fraction of maximum resource capacity required to regen health'''

  RESOURCE_HEALTH_RESTORE_FRACTION    = 0.1
  '''Fraction of health restored per tick when above half food+water'''


class Combat:
  '''Combat Game System'''

  COMBAT_SYSTEM_ENABLED              = True
  '''Game system flag'''

  COMBAT_SPAWN_IMMUNITY              = 20
  '''Agents older than this many ticks cannot attack agents younger than this many ticks'''

  COMBAT_STATUS_DURATION             = 3
  '''Combat status lasts for this many ticks after the last combat event.
     Combat events include both attacking and being attacked.'''

  COMBAT_WEAKNESS_MULTIPLIER         = 1.5
  '''Multiplier for super-effective attacks'''

  def COMBAT_DAMAGE_FORMULA(self, offense, defense, multiplier):
    '''Damage formula'''
    return int(multiplier * (offense * (15 / (15 + defense))))

  COMBAT_MELEE_DAMAGE                = 30
  '''Melee attack damage'''

  COMBAT_MELEE_REACH                 = 3
  '''Reach of attacks using the Melee skill'''

  COMBAT_RANGE_DAMAGE                = 30
  '''Range attack damage'''

  COMBAT_RANGE_REACH                 = 3
  '''Reach of attacks using the Range skill'''

  COMBAT_MAGE_DAMAGE                 = 30
  '''Mage attack damage'''

  COMBAT_MAGE_REACH                  = 3
  '''Reach of attacks using the Mage skill'''


def default_exp_threshold(max_level):
  import math
  additional_exp_per_level = [round(90*math.sqrt(lvl))
                              for lvl in range(1, max_level+1)]
  return [sum(additional_exp_per_level[:lvl]) for lvl in range(max_level)]

class Progression:
  '''Progression Game System'''

  PROGRESSION_SYSTEM_ENABLED        = True
  '''Game system flag'''

  PROGRESSION_BASE_LEVEL            = 1
  '''Initial skill level'''

  PROGRESSION_LEVEL_MAX             = 10
  '''Max skill level'''

  PROGRESSION_EXP_THRESHOLD         = default_exp_threshold(PROGRESSION_LEVEL_MAX)
  '''A list of experience thresholds for each level'''

  PROGRESSION_COMBAT_XP_SCALE       = 3
  '''Additional XP for each attack for skills Melee, Range, and Mage'''

  PROGRESSION_AMMUNITION_XP_SCALE   = 15
  '''Additional XP for each harvest for Prospecting, Carving, and Alchemy'''

  PROGRESSION_CONSUMABLE_XP_SCALE   = 30
  '''Multiplier XP for each harvest for Fishing and Herbalism'''

  PROGRESSION_MELEE_BASE_DAMAGE     = 20
  '''Base Melee attack damage'''

  PROGRESSION_MELEE_LEVEL_DAMAGE    = 5
  '''Bonus Melee attack damage per level'''

  PROGRESSION_RANGE_BASE_DAMAGE     = 20
  '''Base Range attack damage'''

  PROGRESSION_RANGE_LEVEL_DAMAGE    = 5
  '''Bonus Range attack damage per level'''

  PROGRESSION_MAGE_BASE_DAMAGE      = 20
  '''Base Mage attack damage '''

  PROGRESSION_MAGE_LEVEL_DAMAGE     = 5
  '''Bonus Mage attack damage per level'''

  PROGRESSION_BASE_DEFENSE          = 0
  '''Base defense'''

  PROGRESSION_LEVEL_DEFENSE         = 5
  '''Bonus defense per level'''


class NPC:
  '''NPC Game System'''

  NPC_SYSTEM_ENABLED                  = True
  '''Game system flag'''

  NPC_N                               = None
  '''Maximum number of NPCs spawnable in the environment'''

  NPC_SPAWN_ATTEMPTS                  = 25
  '''Number of NPC spawn attempts per tick'''

  NPC_SPAWN_AGGRESSIVE                = 0.80
  '''Percentage distance threshold from spawn for aggressive NPCs'''

  NPC_SPAWN_NEUTRAL                   = 0.50
  '''Percentage distance threshold from spawn for neutral NPCs'''

  NPC_SPAWN_PASSIVE                   = 0.00
  '''Percentage distance threshold from spawn for passive NPCs'''

  NPC_LEVEL_MIN                       = 1
  '''Minimum NPC level'''

  NPC_LEVEL_MAX                       = 10
  '''Maximum NPC level'''

  NPC_BASE_DEFENSE                    = 0
  '''Base NPC defense'''

  NPC_LEVEL_DEFENSE                   = 15
  '''Bonus NPC defense per level'''

  NPC_BASE_DAMAGE                     = 15
  '''Base NPC damage'''

  NPC_LEVEL_DAMAGE                    = 15
  '''Bonus NPC damage per level'''


class Item:
  '''Inventory Game System'''

  ITEM_SYSTEM_ENABLED                 = True
  '''Game system flag'''

  ITEM_N                              = 17
  '''Number of unique base item classes'''

  ITEM_INVENTORY_CAPACITY             = 12
  '''Number of inventory spaces'''

  ITEM_ALLOW_GIFT               = True
  '''Whether agents can give gold/item to each other'''

  @property
  def INVENTORY_N_OBS(self):
    '''Number of distinct item observations'''
    return self.ITEM_INVENTORY_CAPACITY


class Equipment:
  '''Equipment Game System'''

  EQUIPMENT_SYSTEM_ENABLED             = True
  '''Game system flag'''

  WEAPON_DROP_PROB = 0.025
  '''Chance of getting a weapon while harvesting ammunition'''

  EQUIPMENT_WEAPON_BASE_DAMAGE         = 15
  '''Base weapon damage'''

  EQUIPMENT_WEAPON_LEVEL_DAMAGE        = 15
  '''Added weapon damage per level'''

  EQUIPMENT_AMMUNITION_BASE_DAMAGE     = 15
  '''Base ammunition damage'''

  EQUIPMENT_AMMUNITION_LEVEL_DAMAGE    = 15
  '''Added ammunition damage per level'''

  EQUIPMENT_TOOL_BASE_DEFENSE          = 30
  '''Base tool defense'''

  EQUIPMENT_TOOL_LEVEL_DEFENSE         = 0
  '''Added tool defense per level'''

  EQUIPMENT_ARMOR_BASE_DEFENSE         = 0
  '''Base armor defense'''

  EQUIPMENT_ARMOR_LEVEL_DEFENSE        = 10
  '''Base equipment defense'''


class Profession:
  '''Profession Game System'''

  PROFESSION_SYSTEM_ENABLED           = True
  '''Game system flag'''

  PROFESSION_TREE_CAPACITY            = 1
  '''Maximum number of harvests before a tree tile decays'''

  PROFESSION_TREE_RESPAWN             = 0.105
  '''Probability that a harvested tree tile will regenerate each tick'''

  PROFESSION_ORE_CAPACITY             = 1
  '''Maximum number of harvests before an ore tile decays'''

  PROFESSION_ORE_RESPAWN              = 0.10
  '''Probability that a harvested ore tile will regenerate each tick'''

  PROFESSION_CRYSTAL_CAPACITY         = 1
  '''Maximum number of harvests before a crystal tile decays'''

  PROFESSION_CRYSTAL_RESPAWN          = 0.10
  '''Probability that a harvested crystal tile will regenerate each tick'''

  PROFESSION_HERB_CAPACITY            = 1
  '''Maximum number of harvests before an herb tile decays'''

  PROFESSION_HERB_RESPAWN             = 0.02
  '''Probability that a harvested herb tile will regenerate each tick'''

  PROFESSION_FISH_CAPACITY            = 1
  '''Maximum number of harvests before a fish tile decays'''

  PROFESSION_FISH_RESPAWN             = 0.02
  '''Probability that a harvested fish tile will regenerate each tick'''

  @staticmethod
  def PROFESSION_CONSUMABLE_RESTORE(level):
    return 50 + 5*level


class Exchange:
  '''Exchange Game System'''

  EXCHANGE_SYSTEM_ENABLED             = True
  '''Game system flag'''

  EXCHANGE_BASE_GOLD                  = 1
  '''Initial gold amount'''

  EXCHANGE_LISTING_DURATION           = 5
  '''The number of ticks, during which the item is listed for sale'''

  MARKET_N_OBS = 1024
  '''Number of distinct item observations'''

  PRICE_N_OBS = 99 # make it different from PLAYER_N_OBS
  '''Number of distinct price observations
     This also determines the maximum price one can set for an item
  '''


class Communication:
  '''Exchange Game System'''

  COMMUNICATION_SYSTEM_ENABLED             = True
  '''Game system flag'''

  # CHECK ME: When do we actually use this?
  COMMUNICATION_NUM_TOKENS                 = 50
  '''Number of distinct COMM tokens'''


class AllGameSystems(
  Terrain, Resource, Combat, NPC, Progression, Item,
  Equipment, Profession, Exchange, Communication):
  pass


############################################################################
### Config presets
class Small(Config):
  '''A small config for debugging and experiments with an expensive outer loop'''

  PATH_MAPS                    = 'maps/small'

  PLAYER_N                     = 64

  MAP_PREVIEW_DOWNSCALE        = 4
  MAP_CENTER                   = 32

  TERRAIN_LOG_INTERPOLATE_MIN  = 0

  NPC_N                        = 32
  NPC_LEVEL_MAX                = 5
  NPC_LEVEL_SPREAD             = 1

  PROGRESSION_SPAWN_CLUSTERS   = 4
  PROGRESSION_SPAWN_UNIFORMS   = 16

  HORIZON                      = 128


class Medium(Config):
  '''A medium config suitable for most academic-scale research'''

  PATH_MAPS                    = 'maps/medium'

  PLAYER_N                     = 128

  MAP_PREVIEW_DOWNSCALE        = 16
  MAP_CENTER                   = 128

  NPC_N                        = 128
  NPC_LEVEL_MAX                = 10
  NPC_LEVEL_SPREAD             = 1

  PROGRESSION_SPAWN_CLUSTERS   = 64
  PROGRESSION_SPAWN_UNIFORMS   = 256

  HORIZON                      = 1024


class Large(Config):
  '''A large config suitable for large-scale research or fast models'''

  PATH_MAPS                    = 'maps/large'

  PLAYER_N                     = 1024

  MAP_PREVIEW_DOWNSCALE        = 64
  MAP_CENTER                   = 1024

  NPC_N                        = 1024
  NPC_LEVEL_MAX                = 15
  NPC_LEVEL_SPREAD             = 3

  PROGRESSION_SPAWN_CLUSTERS   = 1024
  PROGRESSION_SPAWN_UNIFORMS   = 4096

  HORIZON                 = 8192


class Default(Medium, AllGameSystems):
  pass



from __future__ import annotations

from typing import Dict

from nmmo.core.agent import Agent
from nmmo.entity.player import Player
from nmmo.lib.log import Logger, MilestoneLogger


class LogHelper:
  @staticmethod
  def create(realm) -> LogHelper:
    if realm.config.LOG_ENV:
      return SimpleLogHelper(realm)
    return DummyLogHelper()

class DummyLogHelper(LogHelper):
  def reset(self) -> None:
    pass

  def update(self, dead_players: Dict[int, Player]) -> None:
    pass

  def log_milestone(self, milestone: str, value: float) -> None:
    pass

  def log_event(self, event: str, value: float) -> None:
    pass

class SimpleLogHelper(LogHelper):
  def __init__(self, realm) -> None:
    self.realm = realm
    self.config = realm.config

    self.reset()

  def reset(self):
    self._env_logger    = Logger()
    self._player_logger = Logger()
    self._event_logger  = DummyLogHelper()
    self._milestone_logger = DummyLogHelper()

    if self.config.LOG_EVENTS:
      self._event_logger = Logger()

    if self.config.LOG_MILESTONES:
      self._milestone_logger = MilestoneLogger(self.config.LOG_FILE)

    self._player_stats_funcs = {}
    self._register_player_stats()

  def log_milestone(self, milestone: str, value: float) -> None:
    if self.config.LOG_MILESTONES:
      self._milestone_logger.log(milestone, value)

  def log_event(self, event: str, value: float) -> None:
    if self.config.LOG_EVENTS:
      self._event_logger.log(event, value)

  @property
  def packet(self):
    packet = {'Env': self._env_logger.stats,
              'Player': self._player_logger.stats}

    if self.config.LOG_EVENTS:
      packet['Event'] = self._event_logger.stats
    else:
      packet['Event'] = 'Unavailable: config.LOG_EVENTS = False'

    if self.config.LOG_MILESTONES:
      packet['Milestone'] = self._event_logger.stats
    else:
      packet['Milestone'] = 'Unavailable: config.LOG_MILESTONES = False'

    return packet

  def _register_player_stat(self, name: str, func: callable):
    assert name not in self._player_stats_funcs
    self._player_stats_funcs[name] = func

  def _register_player_stats(self):
    self._register_player_stat('Basic/TimeAlive', lambda player: player.history.time_alive.val)
    # Skills
    if self.config.PROGRESSION_SYSTEM_ENABLED:
      if self.config.COMBAT_SYSTEM_ENABLED:
        self._register_player_stat('Skill/Mage', lambda player: player.skills.mage.level.val)
        self._register_player_stat('Skill/Range', lambda player: player.skills.range.level.val)
        self._register_player_stat('Skill/Melee', lambda player: player.skills.melee.level.val)
      if self.config.PROFESSION_SYSTEM_ENABLED:
        self._register_player_stat('Skill/Fishing', lambda player: player.skills.fishing.level.val)
        self._register_player_stat('Skill/Herbalism',
          lambda player: player.skills.herbalism.level.val)
        self._register_player_stat('Skill/Prospecting',
          lambda player: player.skills.prospecting.level.val)
        self._register_player_stat('Skill/Carving',
          lambda player: player.skills.carving.level.val)
        self._register_player_stat('Skill/Alchemy',
        lambda player: player.skills.alchemy.level.val)
      if self.config.EQUIPMENT_SYSTEM_ENABLED:
        self._register_player_stat('Item/Held-Level',
          lambda player: player.inventory.equipment.held.item.level.val \
            if player.inventory.equipment.held.item else 0)
        self._register_player_stat('Item/Equipment-Total',
          lambda player: player.equipment.total(lambda e: e.level))

    if self.config.EXCHANGE_SYSTEM_ENABLED:
      self._register_player_stat('Exchange/Player-Sells', lambda player: player.sells)
      self._register_player_stat('Exchange/Player-Buys',  lambda player: player.buys)
      self._register_player_stat('Exchange/Player-Wealth',  lambda player: player.gold.val)

    # Item usage
    if self.config.PROFESSION_SYSTEM_ENABLED:
      self._register_player_stat('Item/Ration-Consumed', lambda player: player.ration_consumed)
      self._register_player_stat('Item/Poultice-Consumed', lambda player: player.poultice_consumed)
      self._register_player_stat('Item/Ration-Level', lambda player: player.ration_level_consumed)
      self._register_player_stat('Item/Poultice-Level',
        lambda player: player.poultice_level_consumed)

  def update(self, dead_players: Dict[int, Player]) -> None:
    for player in dead_players.values():
      for key, val in self._player_stats(player).items():
        self._player_logger.log(key, val)

    # TODO: handle env logging

  def _player_stats(self, player: Agent) -> Dict[str, float]:
    stats = {}
    policy = player.policy

    for key, stat_func in self._player_stats_funcs.items():
      stats[f'{key}_{policy}'] = stat_func(player)

    stats['Time_Alive'] = player.history.time_alive.val

    return stats



import unittest

import numpy as np

from nmmo.datastore.numpy_datastore import NumpyTable

# pylint: disable=protected-access
class TestNumpyTable(unittest.TestCase):
  def test_continous_table(self):
    table = NumpyTable(3, 10, np.float32)
    table.update(2, 0, 2.1)
    table.update(2, 1, 2.2)
    table.update(5, 0, 5.1)
    table.update(5, 2, 5.3)
    np.testing.assert_array_equal(
      table.get([1,2,5]),
      np.array([[0, 0, 0], [2.1, 2.2, 0], [5.1, 0, 5.3]], dtype=np.float32)
    )

  def test_discrete_table(self):
    table = NumpyTable(3, 10, np.int32)
    table.update(2, 0, 11)
    table.update(2, 1, 12)
    table.update(5, 0, 51)
    table.update(5, 2, 53)
    np.testing.assert_array_equal(
      table.get([1,2,5]),
      np.array([[0, 0, 0], [11, 12, 0], [51, 0, 53]], dtype=np.int32)
    )

  def test_expand(self):
    table = NumpyTable(3, 10, np.float32)

    table.update(2, 0, 2.1)
    with self.assertRaises(IndexError):
      table.update(10, 0, 10.1)

    table._expand(11)
    table.update(10, 0, 10.1)

    np.testing.assert_array_equal(
      table.get([10, 2]),
      np.array([[10.1, 0, 0], [2.1, 0, 0]], dtype=np.float32)
    )

if __name__ == '__main__':
  unittest.main()



from collections.abc import Mapping
from typing import Dict

from nmmo.entity.entity import Entity
from nmmo.entity.npc import NPC
from nmmo.entity.player import Player
from nmmo.lib import spawn
from nmmo.systems import combat


class EntityGroup(Mapping):
  def __init__(self, realm, np_random):
    self.datastore = realm.datastore
    self.realm = realm
    self.config = realm.config
    self._np_random = np_random

    self.entities: Dict[int, Entity] = {}
    self.dead_this_tick: Dict[int, Entity] = {}

  def __len__(self):
    return len(self.entities)

  def __contains__(self, e):
    return e in self.entities

  def __getitem__(self, key) -> Entity:
    return self.entities[key]

  def __iter__(self) -> Entity:
    yield from self.entities

  def items(self):
    return self.entities.items()

  @property
  def corporeal(self):
    return {**self.entities, **self.dead_this_tick}

  @property
  def packet(self):
    return {k: v.packet() for k, v in self.corporeal.items()}

  def reset(self, np_random):
    self._np_random = np_random # reset the RNG
    for ent in self.entities.values():
      # destroy the items
      if self.config.ITEM_SYSTEM_ENABLED:
        for item in list(ent.inventory.items):
          item.destroy()
      ent.datastore_record.delete()

    self.entities = {}
    self.dead_this_tick = {}

  def spawn(self, entity):
    pos, ent_id = entity.pos, entity.id.val
    self.realm.map.tiles[pos].add_entity(entity)
    self.entities[ent_id] = entity

  def cull(self):
    self.dead_this_tick = {}
    for ent_id in list(self.entities):
      player = self.entities[ent_id]
      if not player.alive:
        r, c  = player.pos
        ent_id = player.ent_id
        self.dead_this_tick[ent_id] = player

        self.realm.map.tiles[r, c].remove_entity(ent_id)

        # destroy the remaining items (of starved/dehydrated players)
        #    of the agents who don't go through receive_damage()
        if self.config.ITEM_SYSTEM_ENABLED:
          for item in list(player.inventory.items):
            item.destroy()

        self.entities[ent_id].datastore_record.delete()
        del self.entities[ent_id]

    return self.dead_this_tick

  def update(self, actions):
    for entity in self.entities.values():
      entity.update(self.realm, actions)


class NPCManager(EntityGroup):
  def __init__(self, realm, np_random):
    super().__init__(realm, np_random)
    self.next_id = -1
    self.spawn_dangers = []

  def reset(self, np_random):
    super().reset(np_random)
    self.next_id = -1
    self.spawn_dangers = []

  def spawn(self):
    config = self.config

    if not config.NPC_SYSTEM_ENABLED:
      return

    for _ in range(config.NPC_SPAWN_ATTEMPTS):
      if len(self.entities) >= config.NPC_N:
        break

      if self.spawn_dangers:
        danger = self.spawn_dangers[-1]
        r, c   = combat.spawn(config, danger, self._np_random)
      else:
        center = config.MAP_CENTER
        border = self.config.MAP_BORDER
        # pylint: disable=unbalanced-tuple-unpacking
        r, c   = self._np_random.integers(border, center+border, 2).tolist()

      npc = NPC.spawn(self.realm, (r, c), self.next_id, self._np_random)
      if npc:
        super().spawn(npc)
        self.next_id -= 1

    if self.spawn_dangers:
      self.spawn_dangers.pop()

  def cull(self):
    for entity in super().cull().values():
      self.spawn_dangers.append(entity.spawn_danger)

    # refill npcs to target config.NPC_N, within config.NPC_SPAWN_ATTEMPTS
    self.spawn()

  def actions(self, realm):
    actions = {}
    for idx, entity in self.entities.items():
      actions[idx] = entity.decide(realm)
    return actions

class PlayerManager(EntityGroup):
  def __init__(self, realm, np_random):
    super().__init__(realm, np_random)
    self.loader_class = self.realm.config.PLAYER_LOADER
    self._agent_loader: spawn.SequentialLoader = None
    self.spawned = None

  def reset(self, np_random):
    super().reset(np_random)
    self._agent_loader = self.loader_class(self.config, self._np_random)
    self.spawned = set()

  def spawn_individual(self, r, c, idx, resilient=False):
    agent = next(self._agent_loader)
    agent = agent(self.config, idx)
    player = Player(self.realm, (r, c), agent, resilient)
    super().spawn(player)
    self.spawned.add(idx)

  def spawn(self):
    # Check and assign the constant heal flag
    resilient_flag = [False] * self.config.PLAYER_N
    if self.config.RESOURCE_SYSTEM_ENABLED:
      num_resilient = round(self.config.RESOURCE_RESILIENT_POPULATION * self.config.PLAYER_N)
      for idx in range(num_resilient):
        resilient_flag[idx] = self.config.RESOURCE_DAMAGE_REDUCTION > 0
      self._np_random.shuffle(resilient_flag)

    # Spawn the players
    idx = 0
    while idx < self.config.PLAYER_N:
      idx += 1
      r, c = self._agent_loader.get_spawn_position(idx)

      if idx in self.entities:
        continue

      if idx in self.spawned:
        continue

      self.spawn_individual(r, c, idx, resilient_flag[idx-1])



from __future__ import annotations
from typing import Dict, Iterable, Tuple, MutableMapping, Set, List
from dataclasses import dataclass, field
from copy import deepcopy
from collections import defaultdict
import weakref

from abc import ABC, abstractmethod
import functools
import numpy as np

from nmmo.core.config import Config
from nmmo.core.realm import Realm
from nmmo.core.observation import Observation
from nmmo.task.group import Group
from nmmo.entity.entity import EntityState
from nmmo.lib.event_log import EventState, ATTACK_COL_MAP, ITEM_COL_MAP, LEVEL_COL_MAP
from nmmo.lib.log import EventCode
from nmmo.systems.item import ItemState
from nmmo.core.tile import TileState

EntityAttr = EntityState.State.attr_name_to_col
EntityAttrKeys = EntityAttr.keys()
EventAttr = EventState.State.attr_name_to_col
ItemAttr = ItemState.State.attr_name_to_col
TileAttr = TileState.State.attr_name_to_col
EventAttr.update(ITEM_COL_MAP)
EventAttr.update(ATTACK_COL_MAP)
EventAttr.update(LEVEL_COL_MAP)

@dataclass(frozen=True) # make gs read-only, except cache_result
class GameState:
  current_tick: int
  config: Config
  spawn_pos: Dict[int, Tuple[int, int]] # ent_id: (row, col) of all spawned agents

  alive_agents: Set[int] # of alive agents' ent_id (for convenience)
  env_obs: Dict[int, Observation] # env passes the obs of only alive agents

  entity_data: np.ndarray # a copied, whole Entity ds table
  entity_index: Dict[int, Iterable] # precomputed index for where_in_1d
  item_data: np.ndarray # a copied, whole Item ds table
  item_index: Dict[int, Iterable]
  event_data: np.ndarray # a copied, whole Event log table
  event_index: Dict[int, Iterable]

  cache_result: MutableMapping # cache for general memoization
  _group_view: List[GroupView] = field(default_factory=list) # cache for GroupView

  # add helper functions below
  @functools.lru_cache
  def entity_or_none(self, ent_id):
    flt_ent = self.entity_data[:, EntityAttr['id']] == ent_id
    if np.any(flt_ent):
      return EntityState.parse_array(self.entity_data[flt_ent][0])
    return None

  def where_in_id(self, data_type, subject: Iterable[int]):
    k = (data_type, subject)
    if k in self.cache_result:
      return self.cache_result[k]

    if data_type == 'entity':
      flt_idx = [row for sbj in subject for row in self.entity_index.get(sbj,[])]
      self.cache_result[k] = self.entity_data[flt_idx]
    if data_type == 'item':
      flt_idx = [row for sbj in subject for row in self.item_index.get(sbj,[])]
      self.cache_result[k] = self.item_data[flt_idx]
    if data_type == 'event':
      flt_idx = [row for sbj in subject for row in self.event_index.get(sbj,[])]
      self.cache_result[k] = self.event_data[flt_idx]
    if data_type in ['entity', 'item', 'event']:
      return self.cache_result[k]

    raise ValueError("data_type must be in entity, item, event")

  def get_subject_view(self, subject: Group):
    new_group_view = GroupView(self, subject)
    self._group_view.append(new_group_view)
    return new_group_view

  def clear_cache(self):
    # clear the cache, so that this object can be garbage collected
    self.entity_or_none.cache_clear()  # pylint: disable=no-member
    self.cache_result.clear()
    self.alive_agents.clear()
    while self._group_view:
      weakref.ref(self._group_view.pop())  # clear the cache

# Wrapper around an iterable datastore
class CachedProperty:
  def __init__(self, func):
    self.func = func
    # Allows the instance keys to be garbage collected
    #   when they are no longer referenced elsewhere
    self.cache = weakref.WeakKeyDictionary()

  def __get__(self, instance, owner):
    if instance is None:
      return self
    if instance not in self.cache:
      self.cache[instance] = self.func(instance)
    return self.cache[instance]

class ArrayView(ABC):
  def __init__(self,
               mapping,
               name: str,
               gs: GameState,
               subject: Group,
               arr: np.ndarray):
    self._mapping = mapping
    self._name = name
    self._gs = gs
    self._subject = subject
    self._hash = hash(subject) ^ hash(name)
    self._arr = arr
    self._cache = self._gs.cache_result

  def __len__(self):
    return len(self._arr)

  @abstractmethod
  def get_attribute(self, attr) -> np.ndarray:
    raise NotImplementedError

  def __getattr__(self, attr) -> np.ndarray:
    k = (self._hash, attr)
    if k in self._cache:
      return self._cache[k]
    v = object.__getattribute__(self, 'get_attribute')(attr)
    self._cache[k] = v
    return v

class ItemView(ArrayView):
  def __init__(self, gs: GameState, subject: Group, arr: np.ndarray):
    super().__init__(ItemAttr, 'item', gs, subject, arr)
    self._mapping = ItemAttr

  def get_attribute(self, attr) -> np.ndarray:
    return self._arr[:, self._mapping[attr]]

class EntityView(ArrayView):
  def __init__(self, gs: GameState, subject: Group, arr: np.ndarray):
    super().__init__(EntityAttr, 'entity', gs, subject, arr)

  def get_attribute(self, attr) -> np.ndarray:
    return self._arr[:, self._mapping[attr]]

class EventView(ArrayView):
  def __init__(self, gs: GameState, subject: Group, arr: np.ndarray):
    super().__init__(EventAttr, 'event', gs, subject, arr)

  def get_attribute(self, attr) -> np.ndarray:
    assert hasattr(EventCode, attr), 'Invalid event code'
    arr = self._arr[np.in1d(self._arr[:, EventAttr['event']],
                                          getattr(EventCode, attr))]
    return EventCodeView(attr, self._gs, self._subject, arr)

class TileView(ArrayView):
  def __init__(self, gs: GameState, subject: Group, arr: np.ndarray):
    super().__init__(TileAttr, 'tile', gs, subject, arr)

  def get_attribute(self, attr) -> np.ndarray:
    return [o[:, self._mapping[attr]]for o in self._arr]

class EventCodeView(ArrayView):
  def __init__(self,
               name: str,
               gs: GameState,
               subject: Group,
               arr: np.ndarray):
    super().__init__(EventAttr, name, gs, subject, arr)

  def get_attribute(self, attr) -> np.ndarray:
    return self._arr[:, self._mapping[attr]]

# Group
class GroupObsView:
  def __init__(self, gs: GameState, subject: Group):
    self._gs = gs

    valid_agents = filter(lambda eid: eid in gs.env_obs,subject.agents)
    self._obs = [gs.env_obs[ent_id] for ent_id in valid_agents]
    self._subject = subject

  @CachedProperty
  def tile(self):
    return TileView(self._gs, self._subject, [o.tiles for o in self._obs])

  def __getattr__(self, attr):
    return [getattr(o, attr) for o in self._obs]

class GroupView:
  def __init__(self, gs: GameState, subject: Group):
    self._gs = gs
    self._subject = subject
    self._subject_hash = hash(subject)

  @CachedProperty
  def obs(self):
    return GroupObsView(self._gs, self._subject)

  @CachedProperty
  def _sbj_ent(self):
    return self._gs.where_in_id('entity', self._subject.agents)

  @CachedProperty
  def entity(self):
    return EntityView(self._gs, self._subject, self._sbj_ent)

  @CachedProperty
  def _sbj_item(self):
    return self._gs.where_in_id('item', self._subject.agents)

  @CachedProperty
  def item(self):
    return ItemView(self._gs, self._subject, self._sbj_item)

  @CachedProperty
  def _sbj_event(self):
    return self._gs.where_in_id('event', self._subject.agents)

  @CachedProperty
  def event(self):
    return EventView(self._gs, self._subject, self._sbj_event)

  def __getattribute__(self, attr):
    if attr in {'_gs','_subject','_sbj_ent','_sbj_item',
    'entity','item','event','obs', '_subject_hash'}:
      return object.__getattribute__(self, attr)

    # Cached optimization
    k = (self._subject_hash, attr)
    cache = self._gs.cache_result
    if k in cache:
      return cache[k]

    try:
      # Get property
      if attr in EntityAttrKeys:
        v = getattr(self.entity, attr)
      else:
        v = object.__getattribute__(self, attr)
      cache[k] = v
      return v
    except AttributeError:
      # View behavior
      return object.__getattribute__(self._gs, attr)

class GameStateGenerator:
  def __init__(self, realm: Realm, config: Config):
    self.config = deepcopy(config)
    self.spawn_pos: Dict[int, Tuple[int, int]] = {}

    for ent_id, ent in realm.players.items():
      self.spawn_pos.update( {ent_id: ent.pos} )

  def generate(self, realm: Realm, env_obs: Dict[int, Observation]) -> GameState:
    # copy the datastore, by running astype
    entity_all = EntityState.Query.table(realm.datastore).copy()
    alive_agents = entity_all[:, EntityAttr["id"]]
    alive_agents = set(alive_agents[alive_agents > 0])
    item_data = ItemState.Query.table(realm.datastore).copy()
    event_data = EventState.Query.table(realm.datastore).copy()
    return GameState(
      current_tick = realm.tick,
      config = self.config,
      spawn_pos = self.spawn_pos,
      alive_agents = alive_agents,
      env_obs = env_obs,
      entity_data = entity_all,
      entity_index = precompute_index(entity_all, EntityAttr["id"]),
      item_data = item_data,
      item_index = precompute_index(item_data, ItemAttr["owner_id"]),
      event_data = event_data,
      event_index = precompute_index(event_data, EventAttr['ent_id']),
      cache_result = {}
    )

def precompute_index(table, id_col):
  index = defaultdict()
  for row, id_ in enumerate(table[:,id_col]):
    if id_ in index:
      index[id_].append(row)
    else:
      index[id_] = [row]
  return index



from typing import Dict, Tuple

from ordered_set import OrderedSet

from nmmo.systems import item as Item
class EquipmentSlot:
  def __init__(self) -> None:
    self.item = None

  def equip(self, item: Item.Item) -> None:
    self.item = item

  def unequip(self) -> None:
    if self.item:
      self.item.equipped.update(0)
    self.item = None

class Equipment:
  def __init__(self):
    self.hat = EquipmentSlot()
    self.top = EquipmentSlot()
    self.bottom = EquipmentSlot()
    self.held = EquipmentSlot()
    self.ammunition = EquipmentSlot()

  def total(self, lambda_getter):
    items = [lambda_getter(e).val for e in self]
    if not items:
      return 0
    return sum(items)

  def __iter__(self):
    for slot in [self.hat, self.top, self.bottom, self.held, self.ammunition]:
      if slot.item is not None:
        yield slot.item

  def conditional_packet(self, packet, slot_name: str, slot: EquipmentSlot):
    if slot.item:
      packet[slot_name] = slot.item.packet

  @property
  def item_level(self):
    return self.total(lambda e: e.level)

  @property
  def melee_attack(self):
    return self.total(lambda e: e.melee_attack)

  @property
  def range_attack(self):
    return self.total(lambda e: e.range_attack)

  @property
  def mage_attack(self):
    return self.total(lambda e: e.mage_attack)

  @property
  def melee_defense(self):
    return self.total(lambda e: e.melee_defense)

  @property
  def range_defense(self):
    return self.total(lambda e: e.range_defense)

  @property
  def mage_defense(self):
    return self.total(lambda e: e.mage_defense)

  @property
  def packet(self):
    packet = {}

    self.conditional_packet(packet, 'hat',        self.hat)
    self.conditional_packet(packet, 'top',        self.top)
    self.conditional_packet(packet, 'bottom',     self.bottom)
    self.conditional_packet(packet, 'held',       self.held)
    self.conditional_packet(packet, 'ammunition', self.ammunition)

    # pylint: disable=R0801
    # Similar lines here and in npc.py
    packet['item_level']    = self.item_level
    packet['melee_attack']  = self.melee_attack
    packet['range_attack']  = self.range_attack
    packet['mage_attack']   = self.mage_attack
    packet['melee_defense'] = self.melee_defense
    packet['range_defense'] = self.range_defense
    packet['mage_defense']  = self.mage_defense

    return packet


class Inventory:
  def __init__(self, realm, entity):
    config           = realm.config
    self.realm       = realm
    self.entity      = entity
    self.config      = config

    self.equipment   = Equipment()
    self.capacity = 0

    if config.ITEM_SYSTEM_ENABLED and entity.is_player:
      self.capacity         = config.ITEM_INVENTORY_CAPACITY

    self._item_stacks: Dict[Tuple, Item.Stack] = {}
    self.items: OrderedSet[Item.Item] = OrderedSet([]) # critical for correct functioning

  @property
  def space(self):
    return self.capacity - len(self.items)

  def has_stack(self, signature: Tuple) -> bool:
    return signature in self._item_stacks

  def packet(self):
    item_packet = []
    if self.config.ITEM_SYSTEM_ENABLED:
      item_packet = [e.packet for e in self.items]

    return {
          'items':     item_packet,
          'equipment': self.equipment.packet}

  def __iter__(self):
    for item in self.items:
      yield item

  def receive(self, item: Item.Item) -> bool:
    # Return True if the item is received
    assert isinstance(item, Item.Item), f'{item} received is not an Item instance'
    assert item not in self.items, f'{item} object received already in inventory'
    assert not item.equipped.val, f'Received equipped item {item}'
    assert not item.listed_price.val, f'Received listed item {item}'
    assert item.quantity.val, f'Received empty item {item}'

    if isinstance(item, Item.Stack):
      signature = item.signature
      if self.has_stack(signature):
        stack = self._item_stacks[signature]
        assert item.level.val == stack.level.val, f'{item} stack level mismatch'
        stack.quantity.increment(item.quantity.val)
        # destroy the original item instance after the transfer is complete
        item.destroy()
        return False

      if not self.space:
        # if no space thus cannot receive, just destroy the item
        item.destroy()
        return False

      self._item_stacks[signature] = item

    if not self.space:
      # if no space thus cannot receive, just destroy the item
      item.destroy()
      return False

    self.realm.log_milestone(f'Receive_{item.__class__.__name__}', item.level.val,
      f'INVENTORY: Received level {item.level.val} {item.__class__.__name__}',
      tags={"player_id": self.entity.ent_id})

    item.owner_id.update(self.entity.id.val)
    self.items.add(item)
    return True

  # pylint: disable=protected-access
  def remove(self, item, quantity=None):
    assert isinstance(item, Item.Item), f'{item} removing item is not an Item instance'
    assert item in self.items, f'No item {item} to remove'

    if isinstance(item, Item.Equipment) and item.equipped.val:
      item.unequip(item._slot(self.entity))

    if isinstance(item, Item.Stack):
      signature = item.signature

      assert self.has_stack(item.signature), f'{item} stack to remove not in inventory'
      stack = self._item_stacks[signature]

      if quantity is None or stack.quantity.val == quantity:
        self._remove(stack)
        del self._item_stacks[signature]
        return

      assert 0 < quantity <= stack.quantity.val, \
        f'Invalid remove {quantity} x {item} ({stack.quantity.val} available)'
      stack.quantity.val -= quantity
      return

    self._remove(item)

  def _remove(self, item):
    self.realm.exchange.unlist_item(item)
    item.owner_id.update(0)
    self.items.remove(item)



from collections import defaultdict

import logging


class Logger:
  def __init__(self):
    self.stats = defaultdict(list)

  def log(self, key, val):
    if not isinstance(val, (int, float)):
      raise RuntimeError(f'{val} must be int or float')

    self.stats[key].append(val)
    return True

class MilestoneLogger(Logger):
  def __init__(self, log_file):
    super().__init__()
    logging.basicConfig(format='%(levelname)s:%(message)s',
      level=logging.INFO, filename=log_file, filemode='w')

  def log_min(self, key, val):
    if key in self.stats and val >= self.stats[key][-1]:
      return False

    self.log(key, val)
    return True

  def log_max(self, key, val):
    if key in self.stats and val <= self.stats[key][-1]:
      return False

    self.log(key, val)
    return True


class EventCode:
  # Move
  EAT_FOOD = 1
  DRINK_WATER = 2
  GO_FARTHEST = 3 # record when breaking the previous record

  # Attack
  SCORE_HIT = 11
  PLAYER_KILL = 12

  # Item
  CONSUME_ITEM = 21
  GIVE_ITEM = 22
  DESTROY_ITEM = 23
  HARVEST_ITEM = 24
  EQUIP_ITEM = 25
  LOOT_ITEM = 26

  # Exchange
  GIVE_GOLD = 31
  LIST_ITEM = 32
  EARN_GOLD = 33
  BUY_ITEM = 34
  #SPEND_GOLD = 35 # BUY_ITEM, price has the same info

  # Level up
  LEVEL_UP = 41



class Agent:
  policy   = 'Neural'

  def __init__(self, config, idx):
    '''Base class for agents

    Args:
      config: A Config object
      idx: Unique AgentID int
    '''
    self.config = config
    self.iden   = idx
    self._np_random = None

  def __call__(self, obs):
    '''Used by scripted agents to compute actions. Override in subclasses.

    Args:
        obs: Agent observation provided by the environment
    '''

  def set_rng(self, np_random):
    '''Set the random number generator for the agent for reproducibility

    Args:
        np_random: A numpy random.Generator object
    '''
    self._np_random = np_random

class Scripted(Agent):
  '''Base class for scripted agents'''
  policy   = 'Scripted'




import math
from types import SimpleNamespace

import numpy as np

from nmmo.core.config import Config
from nmmo.datastore.serialized import SerializedState
from nmmo.systems import inventory
from nmmo.lib.log import EventCode

# pylint: disable=no-member
EntityState = SerializedState.subclass(
  "Entity", [
    "id",
    "npc_type", # 1 - passive, 2 - neutral, 3 - aggressive
    "row",
    "col",

    # Status
    "damage",
    "time_alive",
    "freeze",
    "item_level",
    "attacker_id",
    "latest_combat_tick",
    "message",

    # Resources
    "gold",
    "health",
    "food",
    "water",

    # Combat Skills
    "melee_level",
    "melee_exp",
    "range_level",
    "range_exp",
    "mage_level",
    "mage_exp",

    # Harvest Skills
    "fishing_level",
    "fishing_exp",
    "herbalism_level",
    "herbalism_exp",
    "prospecting_level",
    "prospecting_exp",
    "carving_level",
    "carving_exp",
    "alchemy_level",
    "alchemy_exp",
  ])

EntityState.Limits = lambda config: {
  **{
    "id": (-math.inf, math.inf),
    "npc_type": (0, 4),
    "row": (0, config.MAP_SIZE-1),
    "col": (0, config.MAP_SIZE-1),
    "damage": (0, math.inf),
    "time_alive": (0, math.inf),
    "freeze": (0, 3),
    "item_level": (0, 5*config.NPC_LEVEL_MAX),
    "attacker_id": (-np.inf, math.inf),
    "latest_combat_tick": (0, math.inf),
    "health": (0, config.PLAYER_BASE_HEALTH),
  },
  **({
    "message": (0, config.COMMUNICATION_NUM_TOKENS),
  } if config.COMMUNICATION_SYSTEM_ENABLED else {}),
  **({
    "gold": (0, math.inf),
    "food": (0, config.RESOURCE_BASE),
    "water": (0, config.RESOURCE_BASE),
  } if config.RESOURCE_SYSTEM_ENABLED else {}),
  **({
    "melee_level": (0, config.PROGRESSION_LEVEL_MAX),
    "melee_exp": (0, math.inf),
    "range_level": (0, config.PROGRESSION_LEVEL_MAX),
    "range_exp": (0, math.inf),
    "mage_level": (0, config.PROGRESSION_LEVEL_MAX),
    "mage_exp": (0, math.inf),
    "fishing_level": (0, config.PROGRESSION_LEVEL_MAX),
    "fishing_exp": (0, math.inf),
    "herbalism_level": (0, config.PROGRESSION_LEVEL_MAX),
    "herbalism_exp": (0, math.inf),
    "prospecting_level": (0, config.PROGRESSION_LEVEL_MAX),
    "prospecting_exp": (0, math.inf),
    "carving_level": (0, config.PROGRESSION_LEVEL_MAX),
    "carving_exp": (0, math.inf),
    "alchemy_level": (0, config.PROGRESSION_LEVEL_MAX),
    "alchemy_exp": (0, math.inf),
  } if config.PROGRESSION_SYSTEM_ENABLED else {}),
}

EntityState.Query = SimpleNamespace(
  # Whole table
  table=lambda ds: ds.table("Entity").where_neq(
    EntityState.State.attr_name_to_col["id"], 0),

  # Single entity
  by_id=lambda ds, id: ds.table("Entity").where_eq(
    EntityState.State.attr_name_to_col["id"], id)[0],

  # Multiple entities
  by_ids=lambda ds, ids: ds.table("Entity").where_in(
    EntityState.State.attr_name_to_col["id"], ids),

  # Entities in a radius
  window=lambda ds, r, c, radius: ds.table("Entity").window(
    EntityState.State.attr_name_to_col["row"],
    EntityState.State.attr_name_to_col["col"],
    r, c, radius),
)

class Resources:
  def __init__(self, ent, config):
    self.config = config
    self.health = ent.health
    self.water = ent.water
    self.food = ent.food
    self.health_restore = 0
    self.resilient = False

    self.health.update(config.PLAYER_BASE_HEALTH)
    if config.RESOURCE_SYSTEM_ENABLED:
      self.water.update(config.RESOURCE_BASE)
      self.food.update(config.RESOURCE_BASE)

  def update(self):
    if not self.config.RESOURCE_SYSTEM_ENABLED:
      return

    regen = self.config.RESOURCE_HEALTH_RESTORE_FRACTION
    thresh = self.config.RESOURCE_HEALTH_REGEN_THRESHOLD

    food_thresh = self.food > thresh * self.config.RESOURCE_BASE
    water_thresh = self.water > thresh * self.config.RESOURCE_BASE

    org_health = self.health.val
    if food_thresh and water_thresh:
      restore = np.floor(self.health.max * regen)
      self.health.increment(restore)

    if self.food.empty:
      starvation_damage = self.config.RESOURCE_STARVATION_RATE
      if self.resilient:
        starvation_damage *= self.config.RESOURCE_DAMAGE_REDUCTION
      self.health.decrement(int(starvation_damage))

    if self.water.empty:
      dehydration_damage = self.config.RESOURCE_DEHYDRATION_RATE
      if self.resilient:
        dehydration_damage *= self.config.RESOURCE_DAMAGE_REDUCTION
      self.health.decrement(int(dehydration_damage))

    # records both increase and decrease in health due to food and water
    self.health_restore = self.health.val - org_health

  def packet(self):
    data = {}
    data['health'] = { 'val': self.health.val, 'max': self.config.PLAYER_BASE_HEALTH }
    data['food'] = { 'val': self.food.val, 'max': self.config.RESOURCE_BASE }
    data['water'] = { 'val': self.water.val, 'max': self.config.RESOURCE_BASE }
    return data

class Status:
  def __init__(self, ent):
    self.freeze = ent.freeze

  def update(self):
    if self.freeze.val > 0:
      self.freeze.decrement(1)

  def packet(self):
    data = {}
    data['freeze'] = self.freeze.val
    return data


# NOTE: History.packet() is actively used in visulazing attacks
class History:
  def __init__(self, ent):
    self.actions = {}
    self.attack = None

    self.starting_position = ent.pos
    self.exploration = 0
    self.player_kills = 0

    self.damage_received = 0
    self.damage_inflicted = 0

    self.damage = ent.damage
    self.time_alive = ent.time_alive

    self.last_pos = None

  def update(self, entity, actions):
    self.attack = None
    self.damage.update(0)

    self.actions = {}
    if entity.ent_id in actions:
      self.actions = actions[entity.ent_id]

    self.time_alive.increment()

  def packet(self):
    data = {}
    data['damage'] = self.damage.val
    data['timeAlive'] = self.time_alive.val
    data['damage_inflicted'] = self.damage_inflicted
    data['damage_received'] = self.damage_received

    if self.attack is not None:
      data['attack'] = self.attack

    # NOTE: the client seems to use actions for visualization
    #   but produces errors with the new actions. So we comment out these for now
    # actions = {}
    # for atn, args in self.actions.items():
    #   atn_packet = {}

    #   # Avoid recursive player packet
    #   if atn.__name__ == 'Attack':
    #     continue

    #   for key, val in args.items():
    #     if hasattr(val, 'packet'):
    #       atn_packet[key.__name__] = val.packet
    #     else:
    #       atn_packet[key.__name__] = val.__name__
    #   actions[atn.__name__] = atn_packet
    # data['actions'] = actions
    data['actions'] = {}

    return data

# pylint: disable=no-member
class Entity(EntityState):
  def __init__(self, realm, pos, entity_id, name):
    super().__init__(realm.datastore, EntityState.Limits(realm.config))

    self.realm = realm
    self.config: Config = realm.config
    # TODO: do not access realm._np_random directly
    #   related to the whole NPC, scripted logic
    # pylint: disable=protected-access
    self._np_random = realm._np_random

    self.policy = name
    self.entity_id = entity_id
    self.repr = None

    self.name = name + str(entity_id)

    self.row.update(pos[0])
    self.col.update(pos[1])
    self.id.update(entity_id)

    self.vision = self.config.PLAYER_VISION_RADIUS

    self.attacker = None
    self.target = None
    self.closest = None
    self.spawn_pos = pos

    # Submodules
    self.status = Status(self)
    self.history = History(self)
    self.resources = Resources(self, self.config)
    self.inventory = inventory.Inventory(realm, self)

  @property
  def ent_id(self):
    return self.id.val

  def packet(self):
    data = {}
    data['status'] = self.status.packet()
    data['history'] = self.history.packet()
    data['inventory'] = self.inventory.packet()
    data['alive'] = self.alive
    data['base'] = {
      'r': self.row.val,
      'c': self.col.val,
      'name': self.name,
      'level': self.attack_level,
      'item_level': self.item_level.val,
    }

    return data

  def update(self, realm, actions):
    '''Update occurs after actions, e.g. does not include history'''
    if self.history.damage == 0:
      self.attacker = None
      self.attacker_id.update(0)

    if realm.config.EQUIPMENT_SYSTEM_ENABLED:
      self.item_level.update(self.equipment.total(lambda e: e.level))

    self.status.update()
    self.history.update(self, actions)

  # Returns True if the entity is alive
  def receive_damage(self, source, dmg):
    self.history.damage_received += dmg
    self.history.damage.update(dmg)
    self.resources.health.decrement(dmg)

    if self.alive:
      return True

    # at this point, self is dead
    if source:
      source.history.player_kills += 1
      self.realm.event_log.record(EventCode.PLAYER_KILL, source, target=self)

    # if self is dead, unlist its items from the market regardless of looting
    if self.config.EXCHANGE_SYSTEM_ENABLED:
      for item in list(self.inventory.items):
        self.realm.exchange.unlist_item(item)

    # if self is dead but no one can loot, destroy its items
    if source is None or not source.is_player: # nobody or npcs cannot loot
      if self.config.ITEM_SYSTEM_ENABLED:
        for item in list(self.inventory.items):
          item.destroy()
      return False

    # now, source can loot the dead self
    return False

  # pylint: disable=unused-argument
  def apply_damage(self, dmg, style):
    self.history.damage_inflicted += dmg

  @property
  def pos(self):
    return self.row.val, self.col.val

  @property
  def alive(self):
    if self.resources.health.empty:
      return False

    return True

  @property
  def is_player(self) -> bool:
    return False

  @property
  def is_npc(self) -> bool:
    return False

  @property
  def attack_level(self) -> int:
    melee = self.skills.melee.level.val
    ranged = self.skills.range.level.val
    mage = self.skills.mage.level.val

    return int(max(melee, ranged, mage))

  @property
  def in_combat(self) -> bool:
    # NOTE: the initial latest_combat_tick is 0, and valid values are greater than 0
    if not self.config.COMBAT_SYSTEM_ENABLED or self.latest_combat_tick.val == 0:
      return False

    return (self.realm.tick - self.latest_combat_tick.val) < self.config.COMBAT_STATUS_DURATION



# copied from https://github.com/openai/gym/blob/master/gym/utils/seeding.py

"""Set of random number generator functions: seeding, generator, hashing seeds."""
from typing import Any, Optional, Tuple

import numpy as np

from gym import error


class RandomNumberGenerator(np.random.Generator):
  def __init__(self, bit_generator):
    super().__init__(bit_generator)
    self._dir_seq_len = 1024
    self._wrap = self._dir_seq_len - 1
    self._dir_seq = list(self.integers(0, 4, size=self._dir_seq_len))
    self._dir_idx = 0

  # provide a random direction from the pre-generated sequence
  def get_direction(self):
    self._dir_idx = (self._dir_idx + 1) & self._wrap
    return self._dir_seq[self._dir_idx]

def np_random(seed: Optional[int] = None) -> Tuple[np.random.Generator, Any]:
  """Generates a random number generator from the seed and returns the Generator and seed.

  Args:
      seed: The seed used to create the generator

  Returns:
      The generator and resulting seed

  Raises:
      Error: Seed must be a non-negative integer or omitted
  """
  if seed is not None and not (isinstance(seed, int) and 0 <= seed):
    raise error.Error(f"Seed must be a non-negative integer or omitted, not {seed}")

  seed_seq = np.random.SeedSequence(seed)
  np_seed = seed_seq.entropy
  rng = RandomNumberGenerator(np.random.PCG64(seed_seq))
  return rng, np_seed



# pylint: disable=invalid-name, unused-argument
import numpy as np

import nmmo
from nmmo.core.observation import Observation
from nmmo.entity.entity import EntityState
from nmmo.lib import utils


def closestTarget(config, ob: Observation):
  shortestDist = np.inf
  closestAgent = None

  agent  = ob.agent()
  start = (agent.row, agent.col)

  for target_ent in ob.entities.values:
    target_ent = EntityState.parse_array(target_ent)
    if target_ent.id == agent.id:
      continue

    dist = utils.linf_single(start, (target_ent.row, target_ent.col))
    if dist < shortestDist and dist != 0:
      shortestDist = dist
      closestAgent = target_ent

  if closestAgent is None:
    return None, None

  return closestAgent, shortestDist

def attacker(config, ob: Observation):
  agent = ob.agent()

  attacker_id = agent.attacker_id
  if attacker_id == 0:
    return None, None

  target_ent = ob.entity(attacker_id)
  if target_ent is None:
    return None, None

  return target_ent,\
         utils.linf_single((agent.row, agent.col), (target_ent.row, target_ent.col))

def target(config, actions, style, targetID):
  actions[nmmo.action.Attack] = {
        nmmo.action.Style: style,
        nmmo.action.Target: targetID}



import unittest
from timeit import timeit
import numpy as np
from tqdm import tqdm

import nmmo
from nmmo.lib import seeding
from tests.testhelpers import ScriptedAgentTestConfig, ScriptedAgentTestEnv
from tests.testhelpers import observations_are_equal

# 30 seems to be enough to test variety of agent actions
TEST_HORIZON = 30
RANDOM_SEED = np.random.randint(0, 100000)


class TestDeterminism(unittest.TestCase):
  def test_np_random_get_direction(self):
    # pylint: disable=protected-access,bad-builtin,unnecessary-lambda
    np_random_1, np_seed_1 = seeding.np_random(RANDOM_SEED)
    np_random_2, np_seed_2 = seeding.np_random(RANDOM_SEED)
    self.assertEqual(np_seed_1, np_seed_2)

    # also test get_direction, which was added for speed optimization
    self.assertTrue(np.array_equal(np_random_1._dir_seq, np_random_2._dir_seq))

    print('---test_np_random_get_direction---')
    print('np_random.integers():', timeit(lambda: np_random_1.integers(0,4),
                                          number=100000, globals=globals()))
    print('np_random.get_direction():', timeit(lambda: np_random_1.get_direction(),
                                                number=100000, globals=globals()))

  def test_map_determinism(self):
    config = nmmo.config.Default()
    config.MAP_FORCE_GENERATION = True
    config.TERRAIN_FLIP_SEED = False

    map_generator = config.MAP_GENERATOR(config)
    np_random1, _ = seeding.np_random(RANDOM_SEED)
    np_random1_1, _ = seeding.np_random(RANDOM_SEED)

    terrain1, tiles1 = map_generator.generate_map(0, np_random1)
    terrain1_1, tiles1_1 = map_generator.generate_map(0, np_random1_1)

    self.assertTrue(np.array_equal(terrain1, terrain1_1))
    self.assertTrue(np.array_equal(tiles1, tiles1_1))

    # test flip seed
    config2 = nmmo.config.Default()
    config2.MAP_FORCE_GENERATION = True
    config2.TERRAIN_FLIP_SEED = True

    map_generator2 = config2.MAP_GENERATOR(config2)
    np_random2, _ = seeding.np_random(RANDOM_SEED)
    terrain2, tiles2 = map_generator2.generate_map(0, np_random2)

    self.assertFalse(np.array_equal(terrain1, terrain2))
    self.assertFalse(np.array_equal(tiles1, tiles2))

  def test_env_level_rng(self):
    # two envs running independently should return the same results

    # config to always generate new maps, to test map determinism
    config1 = ScriptedAgentTestConfig()
    setattr(config1, 'MAP_FORCE_GENERATION', True)
    setattr(config1, 'PATH_MAPS', 'maps/det1')
    setattr(config1, 'RESOURCE_RESILIENT_POPULATION', 0.2)  # uses np_random
    config2 = ScriptedAgentTestConfig()
    setattr(config2, 'MAP_FORCE_GENERATION', True)
    setattr(config2, 'PATH_MAPS', 'maps/det2')
    setattr(config2, 'RESOURCE_RESILIENT_POPULATION', 0.2)

    # to create the same maps, seed must be provided
    env1 = ScriptedAgentTestEnv(config1, seed=RANDOM_SEED)
    env2 = ScriptedAgentTestEnv(config2, seed=RANDOM_SEED)
    envs = [env1, env2]

    init_obs = [env.reset(seed=RANDOM_SEED+1) for env in envs]

    self.assertTrue(observations_are_equal(init_obs[0], init_obs[0])) # sanity check
    self.assertTrue(observations_are_equal(init_obs[0], init_obs[1]),
                    f"The multi-env determinism failed. Seed: {RANDOM_SEED}.")

    for _ in tqdm(range(TEST_HORIZON)):
      # step returns a tuple of (obs, rewards, dones, infos)
      step_results = [env.step({}) for env in envs]
      self.assertTrue(observations_are_equal(step_results[0][0], step_results[1][0]),
                      f"The multi-env determinism failed. Seed: {RANDOM_SEED}.")

    event_logs = [env.realm.event_log.get_data() for env in envs]
    self.assertTrue(np.array_equal(event_logs[0], event_logs[1]),
                    f"The multi-env determinism failed. Seed: {RANDOM_SEED}.")


if __name__ == '__main__':
  unittest.main()
